
'''
2 and 2 fit: various fit-formulae between two points
            in a 2-d space. 
'''
from .distributions import *
from .line import *
from copy import deepcopy

class Fit22:

    def __init__(self,ps,direction=[0,1]):
        assert ps.shape == (2,2)
        assert len(direction) == 2 and set(direction) == {0,1}, "invalid direction"
        self.ps = ps
        self.direction = direction
        assert self.ps[self.direction[1],1] > self.ps[self.direction[0],1], "invalid y's"
        self.f = None

class LogFit22(Fit22):
    '''
    structure fits a 
    '''

    def __init__(self,ps,direction=[0,1]):
        super().__init__(ps,direction)
        self.f = self.fit() 
        self.g = self.yfit()

    def fit(self):

        def f(x):
            # ratio of x on x-span
            r1 = abs(x - self.ps[self.direction[0],0]) / abs(self.ps[1,0] - self.ps[0,0])
            real = log(r1 * 9. + 1.) / log(10.)
            return self.ps[self.direction[0],1] + real * abs(self.ps[0,1] - self.ps[1,1])    
        return f

    def yfit(self):

        def g(y):
            p = log(10) * (y - self.ps[self.direction[0],1]) / (abs(self.ps[self.direction[1],1] - self.ps[self.direction[0],1]))
            ep = (e ** p - 1) / 9
            r = ep * abs(self.ps[1,0] - self.ps[0,0])
            x1 = np.round(self.ps[self.direction[0],0] - r,5)
            x2 = np.round(self.ps[self.direction[0],0] + r,5)

            if x1 >= min(self.ps[:,0]) and x1 <= max(self.ps[:,0]):
                return x1

            if x2 >= min(self.ps[:,0]) and x2 <= max(self.ps[:,0]):
                return x2
            raise ValueError("invalid y-input") 

        return g

# case 1
'''
lf22 = LogFit22(np.array([[30.,12.],[60.,80.]]))
function_to_2dplot(generate_2d_data_from_function,\
    lf22.f,[30.,60.],0.1,())
'''

# case 2
'''
lf22 = LogFit22(np.array([[0.,0.],[1.,1.]]))
function_to_2dplot(generate_2d_data_from_function,\
    lf22.f,[0.,1.],0.1,())
'''

# case 3
'''
lf22 = LogFit22(np.array([[60.,12.],[30.,80.]]))
function_to_2dplot(generate_2d_data_from_function,\
    lf22.f,[30.,60.],0.1,())
'''

# case 4
"""
lf22 = LogFit22(np.array([[60.,80.],[30.,12.]]),direction=[1,0])
function_to_2dplot(generate_2d_data_from_function,\
    lf22.f,[30.,60.],0.1,())
"""

class Exp2Fit22(Fit22):
    '''
    structure fits a 
    '''

    def __init__(self,ps,direction=[0,1]):
        super().__init__(ps,direction)
        self.f = self.fit() 
        self.g = self.yfit() 

    def fit(self):

        def f(x):
            # ratio of x on x-span
            r1 = abs(x - self.ps[self.direction[0],0]) / abs(self.ps[1,0] - self.ps[0,0])
            real = r1 ** 2
            return self.ps[self.direction[0],1] + real * abs(self.ps[0,1] - self.ps[1,1])    
        return f

    def yfit(self):

        def g(y):
            m1 = (y - self.ps[self.direction[0],1]) / abs(self.ps[self.direction[0],1] - self.ps[self.direction[1],1])
            m2 = (self.ps[self.direction[0],0] - self.ps[self.direction[1],0]) ** 2
            m = sqrt(m1 * m2)

            x1 = round(self.ps[self.direction[0],0]  + m,5)
            x2 = round(self.ps[self.direction[0],0] - m ,5)

            if x1 >= min(self.ps[:,0]) and x1 <= max(self.ps[:,0]):
                return x1

            if x2 >= min(self.ps[:,0]) and x2 <= max(self.ps[:,0]):
                return x2
            raise ValueError("invalid y-input") 
            
        return g

# case 1
'''
lf22 = Exp2Fit22(np.array([[60.,80.],[30.,12.]]),direction=[1,0])
function_to_2dplot(generate_2d_data_from_function,\
    lf22.f,[30.,60.],0.1,())
'''
#########


class DCurve:

    def __init__(self,fitstruct,activationDirection):
        self.fs = fitstruct
        self.ad = activationDirection

    def x_given_y(self,y):

        if type(self.fs) == Line:
            return self.fs.x_given_y(y)
        return round(self.fs.g(y),5)

    def y_given_x(self,x):
        if type(self.fs) == Line:
            #try:
            return self.fs.y_given_x(round(x,5))
            #except:
            #    return None
        return round(self.fs.f(x),5)

    def __str__(self):
        l = None
        if type(self.fs) == Line: l = "line"
        elif type(self.fs) == LogFit22: l = "logfit22"
        else: l = "exp2fit22"
        p = self.get_point()

        s = "* struct: {} \n".format(l)
        s2 = "* points\n-- {}\n-- {}\n".format(p[0],p[1])
        s3 = "* direction: {}\n".format(self.ad)
        return s + s2 + s3 + "---"

    def get_point(self):
        if type(self.fs) == Line: 
            return deepcopy(self.fs.endpoints)
        return deepcopy(self.fs.ps)

    def point_range(self):
        '''
        [[x-min,x-max],[y-min,y-max]]
        '''
        p = deepcopy(self.fs.endpoints) if type(self.fs) == Line\
            else deepcopy(self.fs.ps)

        x = np.sort(p[:,0])
        y = np.sort(p[:,1])
        return np.array([x,y])

    def in_point_range(self,p):
        pr = self.point_range()
        # case: in y-range?
        if self.ad in {'l','r'}:
            return p[1] >= pr[1,0] and p[1] <= pr[1,1]
        # case: in x-range?
        else:
            return p[0] >= pr[0,0] and p[0] <= pr[0,1]

    def is_ap(self,p):
        '''
        is activation point?

        NOTE: use at discretion, method assumes curve direction is oriented
              by ascending order on the number line, but correct usage 
              could be ascending or descending order.
        '''

        vp = self.value_on_curve(p)
        axis = 1 if self.ad in {'t','d'} else 0

        if self.ad in {'t','r'}:
            return p[axis] <= vp
        return p[axis] >= vp

    def value_on_curve(self,p):
        if self.ad in {'t','b'}:
            return round(self.y_given_x(p[0]),5)
        return round(self.x_given_y(p[1]),5)

    def modulate_fit(self):
        if type(self.fs) == Line:
            return

        if type(self.fs) == LogFit22:
            c = Exp2Fit22(deepcopy(self.fs.ps),self.fs.direction)
        else:
            c = LogFit22(deepcopy(self.fs.ps),self.fs.direction)
        self.fs = c

    def form_point_sequence(self,pointHop = DEFAULT_TRAVELLING_HOP):
        if type(self.fs) == Line:
            self.fs.form_point_sequence(pointHop = pointHop)
            return self.fs.data

        xys = []
        xs = self.point_range()[0]
        i = xs[0]
        while i < xs[1]:
            y = self.y_given_x(i)  
            xys.append((i,y))
            i += pointHop
        return xys



from collections import Counter
from math import factorial
from .numerical_generator import *
from copy import deepcopy

def number_combinations(n,m):
    try:
        return factorial(n) / (factorial(m) * factorial(n - m))
    except:
        return 0

def select_indices_from_sequence(sequence,boolV):
    return np.array([sequence[i] for i,s in enumerate(sequence) if boolV[i]])

class SetMerger:
    '''
    Similar to the frequent item-set mining algorithm, Apriori algorithm. Conducts merging of sets based on their likeness.

    :param sv: vector of sets
    :type sv: list<set<T>>
    '''
    def __init__(self, sv):
        self.sv = sv

    @staticmethod
    def set_difference_score(s1,s2):
        '''
        non-commutative operation b/t two equally-sized sets
        '''
        assert type(s1) == type(s2) and type(s1) is set, "[0] invalid sets"
        assert len(s1) == len(s2), "[1] invalid sets"
        return len(s1 - s2)

    ###################### start: determine if nclosed implication
    @staticmethod
    def correct_number_of_elements(n,m):
        assert n >= m, "invalid args."
        number = 0
        for i in range(1,n):
            number += number_combinations(n - i,m-1)
        return number

    @staticmethod
    def is_nclosed_implication(s,n, checkArgs = True):
        if checkArgs:
            q = np.array([len(s_) for s_ in s])
            assert np.all(q == q[0]), "invalid sequence of sets"

        def log_set(s_):
            counter = counter + Counter(s_)

        # check correct number of elements
        if SetMerger.correct_number_of_elements(n,len(s[0])) > len(s):
            return False

        # check correct number of objects
        counter = Counter()

        for s_ in s:
            counter = counter + Counter(s_)
        counts = np.array(list(counter.values()))
        return np.all(counts == counts[0]) and len(counts) >= n

    ###################### end: determine if nclosed implication

    @staticmethod
    def number_of_sets_at_distance_to_others(s,others,wantedDistance):
        assert type(wantedDistance) is int and wantedDistance >= 0, "invalid wanted distance"
        q = 0
        for o in others:
            d = SetMerger.set_difference_score(s,o)
            if d != wantedDistance: continue
            q += 1
        return q

    ######################## start: methods for merging elements

    @staticmethod
    def merge_elements(s):
        q = set()
        for s_ in s:
            q = q | s_
        return q

    def merges_at_index(self,i,outputSource = False):
        """
        Calculates the merges involving the i'th element of `sv`.

        :return: if `outputSource`: list::(merge),list::(elements that make up the merge) o.w. list::(merge)
        :rtype: (list,list) or list
        """
        n = len(self.sv[i])
        pm = self.possible_merges_at_index(i)
        merges = []
        while len(pm) > 0:
            p = pm.pop(0)
            # determine which merges to add to
            for m in merges:
                if SetMerger.number_of_sets_at_distance_to_others(p,m,1) == len(m) and len(m) < n +1:
                    p_ = deepcopy(m)
                    p_.append(p)
                    merges.append(p_)
            merges.append([self.sv[i],p])

        # determine closed implications
        i = 0
        src = []
        while i < len(merges):
            if not SetMerger.is_nclosed_implication(merges[i],n+ 1):
                merges.pop(i)
            else:
                if outputSource:
                    src.append(merges[i])
                merges[i] = SetMerger.merge_elements(merges[i])
                i += 1

        if len(merges) == 0: return None
        return (merges,src) if outputSource else merges

    def possible_merges_at_index(self,i):
        merges = []
        for j in range(i + 1,len(self.sv)):
            if SetMerger.set_difference_score(self.sv[i],self.sv[j]) == 1:
                merges.append(self.sv[j])
        return merges

    def merge_one(self,saveToVar = False,outputSource = False,verbose = False):
        results = []
        results2 = []
        for i in range(len(self.sv) -1):
            ms = self.merges_at_index(i,outputSource)
            if type(ms) != type(None):
                if outputSource:
                    results.extend(ms[0])
                    results2.extend(ms[1])
                else:
                    results.extend(ms)

        if saveToVar and len(results) > 0:
            self.sv = results

        if verbose:
            print("# length of merges @ : ",len(results))

        return (results,results2) if outputSource else results

    def merge(self,verbose = 0):
        c = 1
        while len(self.merge_one(True,False,c)) > 0:#True:#  > 1:
            c +=1
            continue



from morebs2 import relevance_functions
from morebs2 import search_space_iterator
import unittest
import numpy as np
import operator

'''
python -m morebs2.tests.relevance_functions_test  
'''

# modulo
"""

ispx,modulo

apply f(x) on v -> float q
apply modulo 9 on q -> float q2
q is in range (3,7)?
"""
def sample__RCHF__ISPoly__case_1():

    def qf2(f):
        return f % 9
    kwargs1 = ['nr', qf2]

    def cf1(x,x2):
        return x >= x2[0] and x <= x2[1]
    kwargs2 = ['nr', cf1, [3.0,6.0]]
    return relevance_functions.RCHF__ISPoly(5.0,[kwargs1,kwargs2])

"""
    examples:
    epd:
        - rf := reference (vector)
        - dm := euclidean_point_distance | vector modulo
        - cf := operator
        - dt := operand

    in bounds:
        [a] bool::(range each)
        [b] float::(ratio of columns in range)
        [c] bool([b])
"""
##----------------------------------------------------------
class TestRelevanceFunctions(unittest.TestCase):

    # normalized euclidean point distance, w/ reference?
    def test__rf__euclidean_point_distance(self):
        rf = np.array([8.0,2.3,3.1,4.5,7.1,8.8])
        dm = relevance_functions.euclidean_point_distance
        cf = operator.lt

        # distance threshold is float
        dt = 20.0
        q = relevance_functions.addon_singleton__bool__criteria_distance_from_reference(rf, dm, dt,cf)

        # test the points
        p1 = np.array([6.2,3.1,5.5,4.5,7.1,8.8])
        p2 = np.array([25.5,3.1,5.7,14.5,8.0,9.8])
        assert q(p1), "[0] incorrect case 1"
        assert not q(p2), "[0] incorrect case 2"

        # distance threshold is range
        cf = relevance_functions.lambda_floatin
        dt = (5.0,25.0)
        q = relevance_functions.addon_singleton__bool__criteria_distance_from_reference(rf, dm, dt,cf)
        assert not q(p1), "[1] incorrect case 1"
        assert q(p2), "[1] incorrect case 2"

    def test__RChainHead__euclidean_point_distance__(self):

        rf = np.array([8.0,2.3,3.1,4.5,7.1,8.8])
        dm = relevance_functions.euclidean_point_distance
        cf = operator.lt
        dt = 20.0
        kwargs = ['r', rf,dm,cf,dt]

        rc = relevance_functions.RChainHead()
        rc.add_node_at(kwargs)

        # args. for euclidean_point_distance

        p1 = np.array([6.2,3.1,5.5,4.5,7.1,8.8])
        p2 = np.array([25.5,3.1,5.7,14.5,8.0,9.8])
        assert rc.apply(p1), "[0] incorrect case 1"
        assert not rc.apply(p2), "[0] incorrect case 2"

    def test__RChainHead___in_bounds___(self):

        # case 1: bool::(range each)
        p1 = np.array([6.2,3.1,5.5,4.5,7.1,8.8])

            # subcase: dt is pair
        bounds0 = np.array([[3.1,9]])
        kwargs = ['nr', relevance_functions.lambda_pointinbounds, bounds0]
        rc = relevance_functions.RChainHead()
        rc.add_node_at(kwargs)
        assert rc.apply(p1) == True, "case 1.1: incorrect"

            # subcase: dt is bounds
        bounds = np.array([[5.0,7.0],\
                [3.2,4.1],\
                [5,6],\
                [4.5,4.7],\
                [6.5,7.5],\
                [8.5,9.2]])

        kwargs = ['nr', relevance_functions.lambda_pointinbounds, bounds]
        rc = relevance_functions.RChainHead()
        rc.add_node_at(kwargs)
        assert rc.apply(p1) == False, "case 1.2: incorrect"

        # case 2: float::(number of columns in range)
            # subcase: dt is pair
        kwargs = ['nr', relevance_functions.lambda_countpointsinbounds, bounds0]
        rc = relevance_functions.RChainHead()
        rc.add_node_at(kwargs)
        ##print("X ", rc.apply(p1))
        assert rc.apply(p1) == 6, "case 2.1: incorrect"

            # subcase: dt is pair
        bounds2 = np.array([[3.2,8.8]])
        kwargs = ['nr', relevance_functions.lambda_countpointsinbounds, bounds2]
        rc = relevance_functions.RChainHead()
        rc.add_node_at(kwargs)
        ##print("X ", rc.apply(p1))
        assert rc.apply(p1) == 5, "case 2.2: incorrect"

            # subcase: dt is bounds
        kwargs = ['nr', relevance_functions.lambda_countpointsinbounds,bounds]
        rc = relevance_functions.RChainHead()
        rc.add_node_at(kwargs)
        ##print("X ", rc.apply(p1))
        assert rc.apply(p1) == 5, "case 2.3: incorrect"

    # TODO: refactor into 2 branches
    """
    demonstrates how subvector selector is used
    by their input
    """
    def test__RChainHead___subvectorselector___(self):

        ## case 1: uses function `add_on_sample`
        p1 = np.array([6.2,3.1,5.5,4.5,7.1,8.8])

        def add_on_sample(v_):
            return v_ / 1.5 + 1 > np.mean(p1)

        q = relevance_functions.subvector_selector(add_on_sample,outputType =2)
        kwargs = ['nr', q]

        rc = relevance_functions.RChainHead()
        rc.add_node_at(kwargs)

        q = rc.apply(p1)
        assert relevance_functions.equal_iterables(q,np.array([[5.,8.8]])), "[0] invalid result"

    def test__sample__RCHF__ISPoly__case_1(self):
        q = sample__RCHF__ISPoly__case_1()

        p1 = np.array([3.0,4.0,0.0,11.1])
        p2 = np.array([0.0,14.0,2.0,3.1])

        # 486.1 % 9 = 0.1
        assert q(p1) == False, "incorrect ISPoly case 1.1"

        # 3.1
        assert q(p2) == True, "incorrect ISPoly case 1.1"

    def test__sample_rch_3_with_update(self):
            
        def new_sample_ssi(): 
            bounds = np.array([[-5.,15.],
                                [-10.,25.],
                                [16.,71.],
                                [30.,130.]])
            ssiHop = 5
            columnOrder = [0,1,2,3]
            return search_space_iterator.SearchSpaceIterator(bounds, np.copy(bounds[:,0]), columnOrder, ssiHop,True,0)

        s = new_sample_ssi()
        r = relevance_functions.sample_rch_3_with_update(9,7,14,[2,8])

        ts = []
        for i in range(10):
            v = next(s)
            t = r.apply(v)
            ts.append(t)

        answers = [True,True,True,True,True,True,True,False,True,False]
        assert ts == answers, "incorrect, first 10 samples"

        r.update_rch()

        ts = []
        for i in range(10):
            v = next(s)
            t = r.apply(v)
            ts.append(t) 

        answers2 = [False,False,False,False,False,False,False,False,False,False]
        assert ts == answers2, "incorrect, seconds 10 samples,\nwant {}\ngot {}".format(answers2,ts)



################################

if __name__ == '__main__':
    unittest.main()



from morebs2 import poly_struct
import unittest
import numpy as np

'''
python -m morebs2.tests.poly_struct_test  
'''
class TestPolyClasses(unittest.TestCase):
    
    def test__SPoly__apply(self):

        # poly case 1
        sp = poly_struct.SPoly(np.array([12.0,0.0,3.0,1.0,2.0]))

        #   x1
        v1 = sp.apply(3.0)
        print(v1)
        assert v1 == 999 + 3 + 2, "incorrect SPoly case 1.1"

        v2 = sp.apply(0.0)
        assert v2 == 2.0, "incorrect SPoly case 1.2"

    def test__ISPoly__apply(self):
        s = poly_struct.ISPoly(3.0)
        v1 = s.apply(np.array([12.0,0.0,3.0,1.0,2.0]))
        assert v1 == 999 + 3 + 2, "incorrect case 1.1"

    def test__CEPoly____add__(self):
        x = poly_struct.CEPoly(np.array([[4,5],[3,2],[6,7],[5,0]]))
        x2 = poly_struct.CEPoly(np.array([[10,5],[13,7],[122,0]]))
        assert str(x + x2) == ' 19x^7 + 14x^5 + 3x^2 + 127x^0'

    def test__CEPoly____mul__(self):
        x = poly_struct.CEPoly(np.array([[4,5],[3,2],[6,7],[5,0]]))
        x2 = poly_struct.CEPoly(np.array([[4,1],[3,0]]))
        assert str(x * x2) == " 24x^8 + 18x^7 + 16x^6 + 12x^5 + 12x^3 + 9x^2 + 20x^1 + 15x^0"
        assert str(x * 14) == ' 84x^7 + 56x^5 + 42x^2 + 70x^0'

if __name__ == "__main__":
    unittest.main()
    print()



from .ball_comp_components_test_cases import *
import unittest

"""
python -m morebs2.tests.ball_comp_components_test  
"""
class TestBallCompComponents(unittest.TestCase):

    def test__intersection_ratio(self):

        b1,b2 = sample_ball_pair_1()
        ir1,ir2 = ball_comp_components.intersection_ratio(b1,b2)
        assert ir1 == 0.2 and ir2 == 0.25

        # case 2
        b1,b2 = sample_ball_pair_2()
        ir1,ir2 = ball_comp_components.intersection_ratio(b1,b2)
        assert ir1 == 0.0 and ir2 == 1.0

        b1,b2 = sample_ball_pair_3()
        ir1,ir2 = ball_comp_components.intersection_ratio(b1,b2)
        assert ir1 == 0.3 and ir2 == 0.75
        return

    def test__volume_2intersection_estimate(self):
        # case 1
        b1,b2 = sample_ball_pair_1()
        v = ball_comp_components.volume_2intersection_estimate(b1,b2)
        """
        print("ball volumes")
        print(ball_area(b1.radius,4))
        print(ball_area(b2.radius,4))
        print()
        print(intersection_ratio(b1,b2))
        """
        r1,r2 = ball_comp_components.intersection_ratio(b1,b2)
        q = ball_comp_components.ball_area(b2.radius,4) * r2
        assert v <= q

        # case 2
        b1,b2 = sample_ball_pair_2()
        v = ball_comp_components.volume_2intersection_estimate(b1,b2)
        assert v == ball_comp_components.ball_area(b2.radius,4), "incorrect estimation of volume of intersection"

        # case 3
        b1,b2 = sample_ball_pair_3()
        v = ball_comp_components.volume_2intersection_estimate(b1,b2)
        r1,r2 = ball_comp_components.intersection_ratio(b1,b2)
        q = ball_comp_components.ball_area(b2.radius,4) * r2
        assert v < q, "case 4: incorrect estimation"

        # case 4
        b1,b2 = sample_ball_pair_4()
        v = ball_comp_components.volume_2intersection_estimate(b1,b2)
        assert v == 0.0, "case 4: balls do not intersect"

    def test__Ball____add__(self):
        # case 1
        b1,b2 = sample_ball_pair_1()

        c1 = np.array((10.0,10.0,10.0,10.0))
        q1 = np.empty((6,4))
        for i in range(6):
            p = c1 + i * np.array([0.1,-0.3,-0.2,0.1])
            b1.add_element(p)

        c2 = np.array((17.0,10.0,10.0,10.0))
        q2 = np.empty((6,5))
        for i in range(6):
            p = c2 + i * np.array([0.05,-0.03,-0.2,0.01])
            b2.add_element(p)

        b3 = b1 + b2
        for x in b1.data.newData:
            assert ball_comp_components.euclidean_point_distance(b3.center,x) <= b3.radius, "all points of b1 must be in new ball"
        for x in b2.data.newData:
            assert ball_comp_components.euclidean_point_distance(b3.center,x) <= b3.radius, "all points of b2 must be in new ball"
        assert b3.data.newData.shape == (16,4), "invalid new shape"

        # case 2
        b1,b2 = sample_ball_pair_2()
        b3 = b1 + b2
        assert b3 == b1

if __name__ == '__main__':
    unittest.main()


from .relevance_functions import *

class LabelDeltaPattern:

    def __init__(self, tickLength, deviationPr, deviationRange, labelRange,startLabel):
        assert len(labelRange) == 2 and labelRange[1] >= labelRange[0] 
        assert startLabel >= labelRange[0] and startLabel <= labelRange[1]
                
        self.tickLength = tickLength
        self.deviationPr = deviationPr
        self.deviationRange = deviationRange
        self.labelRange = labelRange
        self.label = startLabel
        self.tick = 1
        return

    def output_label(self):
        self.tick = self.tick % self.tickLength
        if self.tick == 0:
            self.label = self.labelRange[0] + (self.label + 1) % (self.labelRange[1] - self.labelRange[0])
        self.tick += 1

        if random.random() < self.deviationPr:
            return self.deviated_label()
        return self.label

    def deviated_label(self):
        dc = int(random.uniform(self.deviationRange[0],self.deviationRange[1] + 1))
        return self.labelRange[0] + (self.label + dc) % (self.labelRange[1] - self.labelRange[0])

class ModularLabeller:

    def __init__(self,ldp):
        assert type(ldp) == LabelDeltaPattern, "invalid ldp"
        self.ldp = ldp

    def label(self, p):
        return np.array(list(p) + [self.ldp.output_label()])

def rch_modular_labeller_1(ldp):

    ml = ModularLabeller(ldp)

    def qf(v):
        return ml.label(v)

    rch = RChainHead()
    kwargs = ['nr',qf]
    rch.add_node_at(kwargs)
    return rch 


from .ball_comp_components import *
from .set_merger import *

def estimate_n_intersection(intersectionAreas):
    return min(intersectionAreas) * 0.5

class DisjunctionVolumeEstimator:
    """
    Estimates the disjunction of a set of balls based on
    induction using 2-intersection volumes.
    """

    def __init__(self):

        # single-ball volumes: int->float
        self.ballVolumes = {}

        # all 2-intersection volumes
        self.d = {}

        # ball-set intersection volumes: str->float
        self.bv = {}

        self.cache1 = [] # single-ball volume
        self.cache2 = [] # 2-intersection volume

    def log_ball_volume(self,b1):
        prev = self.ballVolumes[b1.idn] if b1.idn in self.ballVolumes else None
        self.ballVolumes[b1.idn] = ball_area(b1.radius,b1.center.shape[0])
        self.cache1.append([b1.idn,prev])

    def log_ball_volume_2intersection(self,b1,b2,updateValue=True):
        k = vector_to_string(sorted([b1.idn,b2.idn]))

        # case: do not update
        if not updateValue and k in self.d:
            return

        # log previous value into cache
        x = None if k not in self.d else self.d[k]
        self.cache2.append([k,x])

        est = volume_2intersection_estimate(b1,b2)
        ###print("$EST FOR {}: {}".format(k,est))
        self.d[k] = est

    def clear_cache(self):
        self.cache1 = []
        self.cache2 = []
        return

    def revert_cache_delta(self,cacheId):
        if cacheId == 1:
            c = self.cache1
            d = self.ballVolumes
        else:
            c = self.cache2
            d = self.d

        while len(c) > 0:
            p = c.pop(0)
            if type(p[1]) == type(None):
                del d[p[0]]
            else:
                d[p[0]] = p[1]

    def revert_cache2_delta(self):
        while len(self.cache2) > 0:
            p = self.cache2.pop(0)
            if type(p[1]) == type(None):
                del self.d[p[0]]
            else:
                self.d[p[0]] = p[1]

    def two_intersection_ball_volume(self,k):
        if k not in self.d: return None
        return self.d[k]

    def target_ball_neighbors(self,bIdn):
        s = set()
        for x in self.d.keys():
            q = string_to_vector(x)
            if bIdn in q:
                s = s | {q[0] if q[0] != bIdn else q[1]}
        return s

    def relevant_2intersections_for_ballset(self,bs):
        twoIntersections = []
        for x in self.d.keys():
            q = string_to_vector(x)
            if q[0] in bs and q[1] in bs:
                twoIntersections.append(set(q))
        return twoIntersections

    def estimate_disjunction_at_target_ball(self,bIdn, verbose,capacity = 500):
        # get 1-intersection volume
        tan = self.target_ball_neighbors(bIdn) | {bIdn}
        q = sum([self.ballVolumes[k] for k in tan])

        # get 2-intersection volumes
        ##ti = self.relevant_2intersections_for_ballset(tan)
        ti = self.relevant_2intersections_for_ballset({bIdn})

        # case: no intersections
        if len(ti) == 0:
            return q

        # minus two-intersection volumes
        q2 = np.sum([self.d[vector_to_string(sorted(t))] for t in ti])
        q -= (q2 * 2)
        j = 3
        c = 1.0
        if verbose:
            print("\t\t{} relevant 2-int".format(len(ti)))
        self.sm = SetMerger(ti)

        self.bv = deepcopy(self.d)
        # alternately add and minus j'th intersection volumes
        x = j if verbose else False
        while True:
            # estimate the j'th disjunction value
            a = self.estimate_disjunction_at_target_ball_(x,capacity)
            if a == 0.0:
                break
            q += (a * j * c)

            # increment the coefficients
            c = -1 * c
            j += 1
            if verbose: print("^ depth @ ", j)
        return q

    def estimate_disjunction_at_target_ball_(self,verbose = False,capacity = None):
        """
        Performs a `SetMerger.merge_one` operation and estimate volumes of new
        intersection sets.

        :param verbose:
        :type verbose: bool
        :param capacity: size capacity for merges, use to prevent memory use error.
        :type capacity: int
        """

        # merge one and collect the new merges and their predecessors
        r1,r2 = self.sm.merge_one(True,True,verbose)

        if len(r1) == 0:
            return 0.0

        if len(r1) >= capacity:
            return 0.0

        # calculate the intersection estimate of each predecessor
        vs = []
        for r in r2:
            iv = self.estimate_int_value(r)
            vs.append(iv)

        self.bv = {}
        q = 0.0
        for (r1_,vs_) in zip(r1,vs):
            k = vector_to_string(sorted(r1_))
            self.bv[k] = vs_
            q += vs_
        return q

    def estimate_int_value(self,iSet,):
        """
        Calculates the intersection-related value for disjunction estimation

        :param iSet: ^^^^^
        :type iSet: iter<vector>
        """

        # collect volumes
        ##print("BV\n", self.bv)
        v = []
        for x in iSet:
            q = vector_to_string(sorted(x))
            v.append(self.bv[q])
        return estimate_n_intersection(v)

    def delete_keyset(self,keySet):
        """
        iterates through keyset and deletes all keys found in keyset from
        `ballVolumes` and `d`

        :param keySet: settias correspondas es keys
        :type keySet: list<int>
        """
        # delete ball volume
        for k in keySet:
            del self.ballVolumes[k]

        # delete 2-intersections
        ks = list(self.d.keys())

        for k in ks:
            q = string_to_vector(k)
            if q[0] in keySet or q[1] in keySet:
                del self.d[k]
        return



from .search_space_iterator_test_cases import *
import unittest

'''
python -m morebs2.tests.search_space_iterator_test  
'''

# TODO: write tests on inverted bounds [1,0]
def test__SearchSpaceIterator__set_hop_pattern_fo():
    HopPattern.DEF_INCREMENT_RATIO = 0.5
    ssi = SearchSpaceIterator_case_1()
    ssi.set_hop_pattern_for_columns()

    for (i,x) in enumerate(ssi.hopPatterns):
        print("hop for ", i)
        print(x.hopDirection)
        print()

class TestSearchSpaceIteratorClass(unittest.TestCase):

    def test__SearchSpaceIterator__next__1(self):

        ssi = SearchSpaceIterator_case_1()

        """
        make sure cycle of correct length
        """
        c = 0
        while True:
            c += 1
            next(ssi)
            #print(next(ssi))
            if ssi.reached_end(): break
        assert c == 8, "c is wrong"

    """
    write data out to file
    `message_data/search_space/ssi_case_2.txt`
    """
    def test__SearchSpaceIterator__next__2(self):

        # case 2
        ssi2 = SearchSpaceIterator_case_2()
        ssi2.cycleOn = True

        # check correct endpoint
        ##print("endpoint: ", ssi2.endpoint)
        assert search_space_iterator.equal_iterables(ssi2.endpoint,np.array([0.8,0.8,0.8])), "invalid endpoint"

        # check correct number of elements in cycle
        # loop entirely
        c = 0
        while True:
            c += 1
            next(ssi2)
            #print(next(ssi2))
            if ssi2.reached_end(): break
        assert c == 125, "incorrect number of elements {}".format(c)

    def test__SearchSpaceIterator__next__3(self):

        ssi3 = SearchSpaceIterator_case_3()

        c = 0
        while c < 1000:
            c += 1
            next(ssi3)
            #print(next(ssi3))
            if ssi3.reached_end(): break
        assert c == 27, "c turns out to be more wrong than right"
        return

    '''
    '''
    def test__SearchSpaceIterator__next__4(self):
        ssi = SearchSpaceIterator_case_4()
        ssi.cycleIs = 1
        ssi.adjust_hop_pattern_heads()

        # print out hop pattern for each
        cycleLengths = []
        for hp in ssi.hopPatterns:
            s = search_space_iterator.cycle_hop_pattern(hp)
            cycleLengths.append(len(s))

        # all are length 7
        cycleLengths = np.array(cycleLengths)
        assert np.all(cycleLengths == cycleLengths[0])
        return


    def test__SearchSpaceIterator__next__9(self): 
        ssi = SearchSpaceIterator_case_9() 
        ans = {16: np.array([0.,0.,0.,0.,2.,2.]),\
        71: np.array([0.,0.,0.,2.,0.,1.]),\
        702: np.array([0.,0.,3.,2.,0.,2.])}

        i = 0 
        while not ssi.reached_end(): 
            q_ = next(ssi) 
            if i in ans: 
                q = ans[i] 
                assert matrix_methods.equal_iterables(q,q_)
            
            i += 1
        refpoint = np.array([1.,2.,3.,5.,4.,6.])
        assert matrix_methods.equal_iterables(ssi.referencePoint,refpoint)
        assert i == 2 * 3 * 4 * 5 * 6 * 7 

#-------------------------------------------------------------

    '''
    checks that the endpoint is correct for SSI case 6.
    '''
    def test__SearchSpaceIterator__CASE_6_endpoint(self):
        ssi = SearchSpaceIterator_case_6()
        sol = np.array([0.,9.,8.,3.,2.5])
        assert search_space_iterator.equal_iterables(ssi.endpoint,sol), "incorrect endpoint for case 6\ngot {}\nwant {}".format(ssi.endpoint,sol)

    # TODO: look into rev__next__
    ###
    def test__SearchSpaceIterator__rev__next__(self):
        ssi = SearchSpaceIterator_case_6()
        ssi.cycleOn = True
        ssi.cycleIs = 0

        q = [np.copy(ssi.referencePoint)]
        c = 0
        while not ssi.reached_end():
            ssi.rev__next__()
            q.append(np.copy(ssi.referencePoint))
            c += 1

            if c == 100:
                break
        assert c == 32, "[0] incorrect number of elements for search space {}".format(c)

        q = np.unique(q,axis = 0)
        assert q.shape[0] == 32 and q.shape[1] == 5, "[0] incorrect shape of q"

        # forward loop
        c = 1
        q = [np.copy(next(ssi))]
        while not ssi.reached_end():
            q.append(np.copy(next(ssi)))

            ##print("V ", next(ssi), "\tF ", ssi.reached_end())
            c += 1
            if c == 100:
                break

        assert c == 32, "[1] incorrect number of elements for search space"
        q = np.unique(q,axis = 0)
        assert q.shape[0] == 32 and q.shape[1] == 5, "[0] incorrect shape of q"

    # FAILS:
    def test__SearchSpaceIterator__rev__next__2(self):
        stat = True

        try:
            ssi = SearchSpaceIterator_case_7()
            stat = not stat
        except:
            pass

        assert stat, "invalid bounds for SSI 7"
        return


    def test__SkewedSearchSpaceIterator__case_1(self):
        sssi = SkewedSearchSpaceIterator_case_1()
        #return
        # check for correct number of elements
        c = 0
        while not sssi.reached_end():
            ##print("C ",c, " : ", next(sssi))
            next(sssi)
            c += 1
        assert c == 243, "incorrect number of elements in cycle {}".format(c)
        return

    def test__SkewedSearchSpaceIterator__case_1__set_point(self):

        sssi = SkewedSearchSpaceIterator_case_1()
        ps = []
        for i in range(10):
            q = next(sssi)
            ##print("Q: ",q)
            ps.append(q)

        sp = np.copy(ps[0])

        # case 1: set point at [3]
        sssi.set_value(sp)
        for i in range(3):
            q = next(sssi)
            assert search_space_iterator.equal_iterables(q,ps[1+i])

        # case 2:
        sssi.set_value(ps[4])
        for i in range(3):
            q = next(sssi)
            assert search_space_iterator.equal_iterables(q,ps[5+i])

    def test__SkewedSearchSpaceIterator__case_3(self):
        sssi = SkewedSearchSpaceIterator_case_3()

        # test for correct hop directions
        q = [0,1,1,0,0]
        for (i,hp) in enumerate(sssi.hopPatterns):
            if q[i]:
                assert hp.hopDirection[0] < 0, "[0] incorrect hop direction for {}".format(i)
            else:
                assert hp.hopDirection[0] > 0, "[1] incorrect hop direction for {}".format(i)

        # iterate for 243 times
        c = 0
        q2 = []
        while not sssi.reached_end():
            ##print(next(sssi))
            q3 = next(sssi)
            q2.append(q3)
            c += 1
        assert c == 243, "incorrect number of elements in cycle"

        q2 = np.array(q2)
        g = np.unique(q2,axis = 1).shape[0]
        assert g == 243, "got {} want {}".format(g,242)

        return

    def test__SearchSpaceIterator__case_1__set_point(self):
        ssi = SearchSpaceIterator_case_1()

        q = []
        for i in range(6):
            x = next(ssi)
            q.append(x)
        s = q[2]
        ssi.set_value(s)

        for i in range(3):
            q_ = next(ssi)
            assert search_space_iterator.equal_iterables(q_,q[3 + i])
        return -1

    def test__SkewedSearchSpaceIterator__case_2(self):
        sssi = SkewedSearchSpaceIterator_case_2()

        # iterate for 243 times
        c = 0
        ##q =
        q = []
        while not sssi.reached_end() and c < 1000:
            q.append(next(sssi))
            c += 1
        assert c == 243, "incorrect number of elements in cycle"
        q = np.array(q)
        assert np.unique(q,axis = 0).shape[0] == q.shape[0], "#1 incorrect number of elements in cycle"

"""
def test__SkewedSearchSpaceIterator__case_x():
    sssi = SkewedSearchSpaceIterator_case_1()

    print("SSSI ")
    print(sssi.iBounds)
    print("EP ")
    print(sssi.ref2)
    print()
    #return -1
    ##return -1

    ##return
    # check for correct number of elements
    c = 0
    ##while not sssi.reached_end():
    for i in range(245):
        print("C ",c, " : ", next(sssi), " end? ", sssi.reached_end())
        ##next(sssi)
        c += 1
    ##assert c == 243, "incorrect number of elements in cycle {}".format(c)

    print("REFBOUNDS")
    print(sssi.referenceBounds)
    print(sssi.ref2)
    print(sssi.endpoint)
    print(sssi.skew)
    return

    # check for correct end
    q = next(sssi)
    assert equal_iterables(q,sssi.referenceBounds[:,1])
    return
"""

################################### TODO: incomplete

# test used for the method right after
def test_x():

    h = HopPattern(0.75, 0.0, 1.0,DIR = 0.5)

    h.headIndex = 1.0
    h.head = 1.0

    print("HOP DIR")
    print(h.hopDirection)
    print()

    for i in range(10):
        print(i, " ", next(h), " | ", h.did_cycle())


def test__SearchSpaceIterator___head_is_correct___():
    # set up the SSI

    # case: head is 0
    ssi = SearchSpaceIterator_case_8(0)

    # iterate through
    c = 0
    #while c < 32:
    while not ssi.reached_end():
        print("{}: {}".format(c,next(ssi)))
        c += 1

    print("bounds: ", ssi.bounds)
    print("start: ", ssi.startPoint)
    print("last element: ", ssi.close_cycle())
    print("next: ", next(ssi))

    # case: head is 1
    print("-------")
    ssi = SearchSpaceIterator_case_8(1)

    c = 0
    while not ssi.reached_end():
        print("{}: {}".format(c,next(ssi)))
        c += 1
    print("bounds: ", ssi.bounds)
    print("start: ", ssi.startPoint)
    print("last element: ", ssi.close_cycle())
    print("next: ", next(ssi))


    return -1

if __name__ == "__main__":
    unittest.main()
    print()



from morebs2 import ball_comp_components
import numpy as np

def sample_ball_pair_1():
    c1 = np.array((10.0,10.0,10.0,10.0))
    c2 = np.array((17.0,10.0,10.0,10.0))
    r1 = 5.0
    r2 = 4.0
    b1 = ball_comp_components.Ball.one_ball_(c1,r1)
    b2 = ball_comp_components.Ball.one_ball_(c2,r2)
    return b1,b2

def sample_ball_pair_2():
    c1 = np.array((10.0,10.0,10.0,10.0))
    c2 = np.array((10.0,8.0,10.0,10.0))
    r1 = 5.0
    r2 = 2.0
    b1 = ball_comp_components.Ball.one_ball_(c1,r1)
    b2 = ball_comp_components.Ball.one_ball_(c2,r2)
    return b1,b2

def sample_ball_pair_3():
    c1 = np.array((10.0,10.0,10.0,10.0))
    c2 = np.array((8.0,8.0,8.0,8.0))
    r1 = 5.0
    r2 = 2.0
    b1 = ball_comp_components.Ball.one_ball_(c1,r1)
    b2 = ball_comp_components.Ball.one_ball_(c2,r2)
    return b1,b2

def sample_ball_pair_4():
    c1 = np.array((10.0,10.0,10.0,10.0))
    c2 = np.array((15.0,15.0,15.0,15.0))
    r1 = 5.0
    r2 = 2.0
    b1 = ball_comp_components.Ball.one_ball_(c1,r1)
    b2 = ball_comp_components.Ball.one_ball_(c2,r2)
    return b1,b2



from .set_merger_test_cases import *
import unittest

#########################################################

'''
python -m morebs2.tests.set_merger_test  
'''
class TestSetMerger(unittest.TestCase):

    def test__SetMerger__is_nclosed_implication(self):#

        s1 = sample_set_sequence_1()
        closed = set_merger.SetMerger.is_nclosed_implication(s1,3)
        assert not closed, "case 1 not closed"

        s2 = sample_set_sequence_2()
        closed = set_merger.SetMerger.is_nclosed_implication(s2,3)
        assert closed, "case 2 closed"

        s11 = sample_set_sequence_11()
        closed = set_merger.SetMerger.is_nclosed_implication(s11,3)
        assert closed, "case 3.1 closed"
        closed = set_merger.SetMerger.is_nclosed_implication(s11,4)
        assert closed, "case 3.2 closed"
        closed = set_merger.SetMerger.is_nclosed_implication(s11,5)
        assert closed, "case 3.3 closed"
        closed = set_merger.SetMerger.is_nclosed_implication(s11,6)
        assert closed, "case 3.4 closed"
        closed = set_merger.SetMerger.is_nclosed_implication(s11,7)
        assert not closed, "case 3.5 not closed"
        closed = set_merger.SetMerger.is_nclosed_implication(s11,8)
        assert not closed, "case 3.6 not closed"

        s12 = sample_set_sequence_12()
        closed = set_merger.SetMerger.is_nclosed_implication(s12,3)
        assert not closed, "case 4 not closed"

        s13 = sample_set_sequence_13()
        closed = set_merger.SetMerger.is_nclosed_implication(s13,2)
        assert closed, "case 6 closed"

    def test__SetMerger__possible_merges_at_index(self):
        s12 = sample_set_sequence_12()
        sm = set_merger.SetMerger(s12)

        q = sm.possible_merges_at_index(0)
        assert q == [{0, 2}, {0, 3}, {0, 4}, {0, 5}, {1, 2}, {1, 3}, {1, 4}, {1, 5}]

        q = sm.possible_merges_at_index(1)
        assert q == [{0, 3},{0,4},{0,5},{1,2},{2,3},{2,4},{2,5}]

    def test__SetMerger__merges_at_index(self):
        s12 = sample_set_sequence_12()
        sm = set_merger.SetMerger(s12)
        merges = sm.merges_at_index(True)
        assert merges == [{0, 2, 3}, {0, 2, 4}, {0, 2, 5}]

    def test__SetMerger__merge_one(self):
        s12 = sample_set_sequence_12()
        sm = set_merger.SetMerger(s12)
        sm.merge_one(True)
        assert sm.sv == [{0, 1, 2}, {0, 1, 3}, {0, 1, 4}, {0, 1, 5},\
            {0, 2, 3}, {0, 2, 4}, {0, 2, 5}, {0, 3, 4}, {0, 3, 5},\
            {0, 4, 5}, {1, 2, 3}, {1, 2, 4}, {1, 2, 5}, {1, 3, 4},\
            {1, 3, 5}, {1, 4, 5}, {2, 3, 4}, {2, 3, 5}, {2, 4, 5},\
            {3, 4, 5}, {5, 6, 7}, {5, 6, 8}, {5, 7, 8}, {6, 7, 8}]

        sm.merge_one(True)
        assert sm.sv == [{0, 1, 2, 3}, {0, 1, 2, 4}, {0, 1, 2, 5},\
                    {0, 1, 3, 4}, {0, 1, 3, 5}, {0, 1, 4, 5}, {0, 2, 3, 4},\
                    {0, 2, 3, 5}, {0, 2, 4, 5}, {0, 3, 4, 5}, {1, 2, 3, 4},\
                    {1, 2, 3, 5}, {1, 2, 4, 5}, {1, 3, 4, 5}, {2, 3, 4, 5}, {5, 6, 7, 8}]


    def test__SetMerger__merge(self):
        sm = set_merger.SetMerger(sample_set_sequence_12())
        sm.merge()
        assert sm.sv == [{0, 1, 2, 3, 4, 5}]
    

if __name__ == '__main__':
    unittest.main()



from .numerical_space_data_generator import *
from .ball_comp_components import *
import pickle

class BallClumpDataGenerator:
    """
    generates points of n-dimensional balls based on parameters that detail the bounds
    of the balls, their radii, and then writes those points into a file 

    :param bounds: proper bounds array
    :type bounds: np.ndarray
    :param ballRadii: sequence of ball radii required
    :type ballRadii: iter
    :param clusterInfo: 2-tuple, [0] is sequence of i'th ball's cluster radii ratios, and [1] is sequence of i'th ball's density measures d, each cluster will have d points
    :type clusterInfo: np.ndarray
    :param filePath: save filepath
    :type filePath: str
    """

    def __init__(self,bounds,ballRadii,clusterInfo,filePath):
        self.bounds = np.round(bounds,5) # target bounds
        self.rbounds = None # refitted bounds
        assert len(ballRadii) == len(clusterInfo[0]), "invalid cluster arg. 0"
        assert len(ballRadii) == len(clusterInfo[1]), "invalid cluster arg. 1"

        self.ballRadii = ballRadii
        self.clusterInfo = clusterInfo
        self.clusterInfo_ = []
        self.filePath = filePath

    def set_frame(self):
        """
        Constructs a sequence of balls, and assigns points to them based on `clusterInfo`.

        NOTE: points that may exist in multiple balls will be assigned to only one ball.
        """
        self.balls = BallClumpDataGenerator.ball_clump_frame_generator_1(self.bounds,deepcopy(self.ballRadii))
        self.rbounds = BallClumpDataGenerator.bounds_for_ball_set(self.balls)

        # make cluster frames for each ball
        for (i,b) in enumerate(self.balls):
            bci = BallClumpDataGenerator.ball_data_cluster_frame(b,self.clusterInfo[0][i])
            self.clusterInfo_.append(bci)

        # make points for each ball
        for (i,b) in enumerate(self.balls):

            centers = self.clusterInfo_[i][0]
            radii = self.clusterInfo_[i][1]

            for j, (c,r) in enumerate(zip(centers,radii)):
                d = self.clusterInfo[1][i][j]
                BallClumpDataGenerator.cluster_points_to_ball(b,c,r,d)


    def make_data(self):
        """
        Constructs the dataset by popping one random element in one random ball
        until all Balls are empty.
        """
        def pop_one_random():
            i = random.randrange(len(aliveBalls))
            bi = aliveBalls[i]

            j = random.randrange(self.balls[bi].data.newData.shape[0])
            p = self.balls[bi].data.newData[j]
            self.balls[bi].data.newData = np.concatenate((self.balls[bi].data.newData[:j],\
                                    self.balls[bi].data.newData[j + 1:]))
            if self.balls[bi].data.newData.shape[0] == 0:
                aliveBalls.pop(i)
            return p

        aliveBalls = [i for i in range(len(self.balls))]
        fp = open(self.filePath,'w')
        while len(aliveBalls) > 0:
            q_ = pop_one_random()
            q = vector_to_string(q_,cr)
            fp.write(q + "\n")

        fp.close()

    @staticmethod
    def ballclump_frame_filepath(fp):
        x2 = fp[::-1].find("/")

        # case: no directories
        if x2 == -1:
            x1 = fp.find(".")
            r = fp[:x1]
            r += "__frame" + fp[x1:]
        else:
            r1,r2 = fp[::-1][:x2][::-1], fp[::-1][x2:][::-1]
            x1 = r1.find(".")
            r1 = r1[:x1] + "__frame" + r1[x1:]
            r = r2 + r1
        return r

    def save_frame(self):
        """
        saves instance's data in the form of a list with the variables
        -- rbounds, ball centers, ball radius, cluster centers, cluster radii, cluster densities
        """
        fp = BallClumpDataGenerator.ballclump_frame_filepath(self.filePath)

        ballCenters = [b.center for b in self.balls]
        ballRadii = [b.radius for b in self.balls]
        clusterCenters = [c[0] for c in self.clusterInfo_]
        clusterRadii = [c[1] for c in self.clusterInfo_]
        clusterDensities = self.clusterInfo[1]

        q = [self.rbounds,ballCenters,ballRadii,clusterCenters,\
            clusterRadii,clusterDensities]

        fobj = open(fp,"wb")
        pickle.dump(q,fobj)
        fobj.close()

    @staticmethod
    def ball_clump_frame_generator_1(bounds,ballRadii):
        """
        Calculates a vector of ball centers corresponding to a connected ball set

        :param bounds: proper bounds array
        :type bounds: np.ndarray
        :param ballRadii: sequence of ball radii required
        :type ballRadii: iter
        :return: listia ballco
        :rtype: list
        """

        """
        connects random ball with one present in the running solution
        at a random intersection ratio
        """
        def connect_random_ball(ballRadius):
            i = random.randrange(len(balls))
            maxDistance = ballRadius + balls[i].radius
            randomIntersectionRatio = random.uniform(0.1,0.75)
            q = randomIntersectionRatio * np.min([ballRadius,balls[i].radius])
            maxDistance -= q
            p2 = random_npoint_from_point(balls[i].center,maxDistance)
            b2 = Ball.one_ball_(p2,ballRadius)
            balls.append(b2)

        # choose a random vertex in bounds
        start = random_bounds_edge(bounds)
        c1 = random_npoint_from_point_in_bounds(bounds,start,ballRadii[0])
        r = round(ballRadii.pop(0),5)
        b1 = Ball.one_ball_(c1,ballRadii[0])
        balls = [b1]

        while len(ballRadii) > 0:
            r = round(ballRadii.pop(0),5)
            connect_random_ball(r)
        return balls

    """
    return:
    - [0]: np.ndarray, cluster centers
    - [1]: list, radius
    """
    @staticmethod
    def ball_data_cluster_frame(ball,clusterRadiiRatios):
        """
        Calculates a vector of ball centers corresponding to a connected ball set

        :param bounds: proper bounds array
        :type bounds: np.ndarray
        :param ballRadii: sequence of ball radii required
        :type ballRadii: iter
        :return: (array<centerios pointist>,list<radius>)
        :rtype: (ICE)
        """

        ci1 = np.empty((0,ball.center.shape[0]))
        ci2 = []
        for r in clusterRadiiRatios:
            assert r > 0 and r <= 1.0, "invalid ratio"
            r_ = r * ball.radius

            # calculate max distance from .center
            maxDistance = ball.radius - r_
            distance = random.uniform(0.0,maxDistance)
            p = random_npoint_from_point(ball.center,distance)
            ci1 = np.vstack((ci1,p))
            ci2.append(r_)
        return (ci1,ci2)

    # CAUTION: clusterCenter not checked for existing in Ball
    """
    clusterCenter := vector, lies in `ball`
    clusterDensity := int, number of points to add to cluster
    """
    @staticmethod
    def cluster_points_to_ball(ball,clusterCenter,clusterRadii,clusterDensity):
        for i in range(clusterDensity):
            p = random_npoint_from_point(clusterCenter,clusterRadii)
            ball.add_element(p)

    # CAUTION: not tested!
    # CAUTION: `bl` not checked for dimensions
    @staticmethod
    def bounds_for_ball_set(bl):
        assert len(bl) >0, "invalid ballset"

        def extremum_at_index(i):
            x = [(b.center[i] - b.radius,b.center[i] + b.radius)\
                for b in bl]
            x = np.array(x)
            minimum,maximum = np.min(x[:,0]), np.max(x[:,1])
            return (minimum,maximum)

        q = bl[0].center.shape[0]
        bounds = []
        for i in range(q):
            bounds.append(extremum_at_index(i))
        return np.round(np.array(bounds),5)

    @staticmethod
    def ball_set_in_bounds(bounds,bl):
        for b in bl:
            if not b.in_bounds(bounds): return False
        return True

    @staticmethod
    def ball_subset_size_in_bounds(bounds,bl):
        c = 0
        for b in bl:
            if b.in_bounds(bounds):
                c += 1
        return c

    @staticmethod
    def ballset_is_connected(bl):
        assert len(bl) > 0, "invalid ball-set"

        # NOTE: inefficient
        def neighbors_of_ball(bi):
            neighbors = []
            for i in range(len(bl)):
                if i == bi: continue
                if bl[bi].is_neighbor(bl[i]):
                    neighbors.append(i)
            return neighbors

        connected = set()
        cache = [0]
        while len(cache) > 0:
            c = cache.pop(0)
            if c in connected: continue
            connected |= {c}
            ns = neighbors_of_ball(c)
            cache.extend(ns)
        return len(connected) == len(bl)

    # RE-WRITE THIS
    """
    @staticmethod
    def make_relevance_function(clusterInfo):

        def cf(v):
            for (i,c) in enumerate(clusterInfo[0]):
                #print("euclid's distance ", euclidean_point_distance(c,v))
                #print("cluster radius ", clusterInfo[1][i])
                if euclidean_point_distance(c,v) <= clusterInfo[1][i]:
                    return True
            return False

        rch = RChainHead()
        kwargs = ['nr',cf]
        rch.add_node_at(kwargs)
        return rch
    """

    @staticmethod
    def load_data_by_frame(frame,fp):
        return "list<balls>"



from .rssi_test_cases import *
import unittest
################################ start tests: relevance zoom, euclidean point distance on 5.0
#

'''
python -m morebs2.tests.rssi_test  
'''
class TestRSSIClass(unittest.TestCase):

    """
    >
    """
    def test__rssi__case1(self):
        print("\n\n\t*** testing case 1\n")
        r = sample_rssi_1(rchl = 1)
        rssi.rssi__display_n_bounds(r, 3)
        return

    """
    <
    """
    def test__rssi__case2(self):
        print("\n\n\t*** testing case 2\n")
        r = sample_rssi_1(rchl = 2)
        rssi.rssi__display_n_bounds(r, 2)
        return

    """
    >= and <=
    """
    def test__rssi__case3(self):
        print("\n\n\t*** testing case 3\n")
        r = sample_rssi_1(rchl = 3)
        rssi.rssi__display_n_bounds(r, 2)
        return

        ################################ end tests: relevance zoom, euclidean point distance on 5.0

        ################################ start tests: png, euclidean point distance on 5.0

    def test__rssi__case4(self):
        print("\n\n\t*** testing case 4\n")

        r = sample_rssi_1(rmMode = "prg")
        rssi.rssi__display_n_bounds(r,1)

        # check relevant points
        '''
        print("relevant points")
        for (i,v) in enumerate(rssi.ri.centerResplat.relevantPointsCache):
            print("{} : {}".format(i,v))
        '''
        assert len(r.ri.centerResplat.relevantPointsCache) == 32, "incorrect number of relevant pts."

        #print("\n\n\tNEZXTING 10\n")
        for i in range(10):
            #print(next(rssi))
            q = next(r)
            assert rssi.point_in_bounds(r.ri.centerResplat.centerBounds,q), "point w/ noise must stay in bounds"

        # display index counter
        print("RPI COUNTER")
        print(r.ri.centerResplat.rpIndexCounter)
        assert len(r.ri.centerResplat.rpIndexCounter) > 0, "center-resplat index counter cannot be empty"
        return

# testing for relevant points
################################ end tests: png, euclidean point distance on 5.0

if __name__ == "__main__":
    unittest.main()
    print()


# NEXT: verbose mode



from .globalls import *
import numpy as np
import math

def equal_iterables(i1, i2, roundPlaces = 5):
    """
    equal_iterables

    :param i1: iterable
    :type i1: iter
    :param i2: iterable
    :type i2: iter
    :param roundPlaces: positive int.
    :type roundPlaces: int
    :return: equal_iterables
    :rtype: bool
    """

    if len(i1) != len(i2): return False

    """
    if np.all(np.equal(np.round(i1, roundPlaces), np.round(i2, roundPlaces)) == True): return True
    return False
    """
    q = np.abs(np.array(i1) - np.array(i2))
    return np.all(q <= 10 ** -roundPlaces)

def indices_of_vector_in_matrix(a, v):
    assert is_2dmatrix(a) and is_vector(v), "invalid criteria"
    a = np.round(a,5)
    v = np.round(v,5)
    indices = []
    for (i, x) in enumerate(a):
        if np.all(np.equal(x, v) == True): indices.append(i)
    return np.array(indices)

def is_2dmatrix(m):
    if type(m) is not np.ndarray: return False
    if len(m.shape) != 2: return False
    return True

def is_bounds_vector(b):
    q = is_2dmatrix(b)
    if not q: return q
    return b.shape[1] == 2

def is_vector(m):
    if type(m) is not np.ndarray: return False
    if len(m.shape) != 1: return False
    return True

def is_binary_vector(b):
    assert is_vector(b), "invalid vector"
    q = np.unique(b)
    return q, len(q) == 2

def is_valid_point(point):
    assert not type(point) is np.ndarray, "point cannot be np.ndarray"
    if len(point) != 2: return False
    if not type(point[0]) in [int, float, np.int64, np.float64]: return False# or type(point[0]) is float): return False
    if not type(point[1]) in [int, float, np.int64, np.float64]: return False
    return True

def is_valid_range(r,is_int:bool=True,inclusive:bool=True):
    if not (type(r) in [list,tuple]): return False
    if not (len(r) == 2): return False

    if is_int: 
        if not (type(r[0]) == type(r[1])): return False
        if not (type(r[0]) == int): return False

    if inclusive:
        if not (r[0] <= r[1]): return False
    else: 
        if not (r[0] < r[1]): return False
    return True 

def np_array_to_string(a):
    assert type(a) is np.ndarray, "not np.ndarray"
    s = str(a).replace("\n", "\t")
    return s

# TODO: expand on criteria
# TODO: unused
def frequency_count_over_2dmatrix(a, floatValue, rowOrColumn = None):
    r,c = np.where(a == floatValue)
    if rowOrColumn == "r": return r
    elif rowOrColumn == "c": return c
    return r,c

# TODO: test
def range_intersection(range1,range2):
    assert is_valid_range(range1,is_int=False,inclusive=True)
    assert is_valid_range(range2,is_int=False,inclusive=True)

    # check range 1
    if range1[0] >= range2[0] and range1[0] <= range2[1]: return True
    if range1[1] >= range2[0] and range1[1] <= range2[1]: return True

    # check range 2
    if range2[0] >= range1[0] and range2[0] <= range1[1]: return True
    if range2[1] >= range1[0] and range2[1] <= range1[1]: return True

    return False

# TODO: test
def complement_of_range_in_range(range1, range2):
    assert range1[0] <= range1[1], "invalid range 1"
    if range2 == None: return range1

    assert range2[0] <= range2[1], "invalid range 2"
    assert range2[0] >= range1[0] and range2[1] <= range1[1], "range 2 not in range 1"

    filter_null_range = lambda r: False if abs(r[0] - r[1])\
                                <= 10 ** -5 else True

    q1 = (range1[0], range2[0])
    q2 = (range2[1], range1[1])
    output = []

    if filter_null_range(q1): output.append(q1)
    if filter_null_range(q2): output.append(q2)

    return output

'''
- 2dmatrix, sequence of values, partition centers
'''
def central_sequence_of_sequence(ps):
    assert type(ps) is np.ndarray, "invalid partition sequence"

    cs = np.empty((ps.shape[0] - 1, ps.shape[1]))
    for i in range(1, len(ps)):
        center = (ps[i] + ps[i - 1]) / 2.0
        cs[i - 1] = center
    return np.round(cs,5)

################################################ START: area measures
def is_valid_area(area):
    """
    Area is 2 x 2 matrix in which [0] represents bottom left, [1] represents upper right.
    """
    return False if not is_2dmatrix(area) else True
    if area.shape[0] != 2 or area.shape[1] != 2: return False
    if area[0,0] > area[1,0]: return False
    if area[0,1] > area[1,1]: return False
    return True

def dim_of_area(area):
    assert is_valid_area(area), "invalid area"
    return area[1,0] - area[0,0], area[1,1] - area[0,1]

def value_of_area(area):
    dim = dim_of_area(area)
    return dim[0] * dim[1]

def area_in_area(a1, a2):
    """
is a1 in a2
    """
    assert is_valid_area(a1) and is_valid_area(a2), "input are not areas"
    if not (a1[0,0] >= a2[0,0] and a1[1,0] <= a2[1,0]): return False
    if not (a1[0,1] >= a2[0,1] and a1[1,1] <= a2[1,1]): return False
    return True

def area_intersects(a1, a2):
    assert is_valid_area(a1) and is_valid_area(a2), "input are not areas"

    # left bottom
        # intersects
    # bottom left
    if a1[0,0] >= a2[0,0] and a1[0,0] <= a2[1,0]\
        and a1[0,1] >= a2[0,1] and a1[0,1] <= a2[1,1]:
        return True

    # upper left
    if a1[0,0] >= a2[0,0] and a1[0,0] <= a2[1,0]\
        and a1[1,1] >= a2[0,1] and a1[1,1] <= a2[1,1]:
        return True

    # bottom right
    if a1[1,0] >= a2[0,0] and a1[1,0] <= a2[1,0]\
        and a1[0,1] >= a2[0,1] and a1[0,1] <= a2[1,1]:
        return True

    # upper right
    if a1[1,0] >= a2[0,0] and a1[1,0] <= a2[1,0]\
        and a1[1,1] >= a2[0,1] and a1[1,1] <= a2[1,1]:
        return True
    return False

# TODO: necessary?
def is_area_qualified(unqualified, area):
    for x in unqualified:
        if area_intersects(x, area): return False
    return True

# TODO: test
def point_in_area(point, area):
    assert is_valid_point(point)
    if not (point[0] >= area[0,0] and point[0] <= area[1,0]):
        return False
    if not (point[1] >= area[0,1] and point[1] <= area[1,1]):
        return False
    return True

def is_point_qualified(unqualifiedAreas, point):

    for x in unqualifiedAreas:
        if point_in_area(point, x): return False
    return True

"""
if degree in [0,90,180,270],

0: 0
90: 1
180: 2
270: 3
"""
def quadrant_of_corner_point_in_area(cp, area):
    corners = area_to_corners(area)
    indices = indices_of_vector_in_matrix(corners, cp)
    if len(indices) != 1: raise ValueError("quadrant for corner point {} could not be obtained for\n{}".format(cp, corners))
    return indices[0]

############################################ END: area

############################################ START: qualifying area search

"""
incrementInfo: dict, keys are [str::"axis" = x|y, float::"increment"]
"""
def extreme_xy_from_point(unqualifiedAreas, area, point, incrementInfo):
    # set up increment and qualifying methods for point increment
    if incrementInfo["axis"] == "x":
        if incrementInfo["increment"] >= 0: qf = lambda p: True if p[0] <= area[1,0] else False
        else: qf = lambda p: True if p[0] >= area[0,0] else False
        ifunc = lambda p: (p[0] + incrementInfo["increment"], p[1])
    elif incrementInfo["axis"] == "y":
        if incrementInfo["increment"] >= 0: qf = lambda p: True if p[1] <= area[1,1] else False
        else: qf = lambda p: True if p[1] >= area[1,0] else False
        ifunc = lambda p: (p[0], incrementInfo["increment"] + p[1])
    else:
        raise ValueError("invalid axis")

    # increment
    prev = None
    pp = point
    while qf(pp):
        if not is_point_qualified(unqualifiedAreas, pp):
            break
        prev = pp
        pp = ifunc(pp)

    return prev

# TODO: implement this.
def greatest_qualifying_area_in_vector_direction(vector, unqualifiedAreas, area, incrementDirections, areaWanted):
    return -1

# TODO: areaWanted needs to be tested.
"""
description:
- approximately determines the greatest quadrilateral with point as one of the four corner points.
  Given the start `point`, determines the extreme x and extreme y point using `incrementDirections`.

arguments:
- point: iterable, length 2
- unqualifiedAreas: np.ndarray, n x 2 x 2
- area: np.ndarray, 2 x 2
- incrementDirections: dict, (increment::float), (direction::((0&1)::(1|-1)))
- areaWanted: max|xmax|ymax
"""
def greatest_qualifying_area_in_direction(point, unqualifiedAreas, area, incrementDirections, areaWanted):

    assert type(incrementDirections["increment"]) is float and incrementDirections["increment"] > 0, "inc. wrong type"
    assert incrementDirections["direction"][0] in [-1,1], "axis 0 inc. wrong"
    assert incrementDirections["direction"][1] in [-1,1], "axis 1 inc. wrong"

    # extreme x
    id1 = {"axis": 0}
    id1["increment"] = incrementDirections["increment"] * incrementDirections["direction"][0]
    pointX = greatest_qualifying_fourway_point_at_point(point, unqualifiedAreas, area, id1)

    # extreme y
    id2 = {"axis": 1}
    id2["increment"] = incrementDirections["increment"] * incrementDirections["direction"][1]
    pointY = greatest_qualifying_fourway_point_at_point(point, unqualifiedAreas, area, id2)

    # TODO: refactor below.
    if areaWanted == "max":
        # get other extremes
        maxYGivenX = greatest_qualifying_fourway_point_at_point(pointX, unqualifiedAreas, area, id2)
        maxXGivenY = greatest_qualifying_fourway_point_at_point(pointY, unqualifiedAreas, area, id1)

        # get areas for each
        #   [0]
        mp1 = missing_area_point_for_three_points(np.array([maxYGivenX, pointX, point]))
        area1 = trim_area(area, corners_to_area(np.array([maxYGivenX, pointX, point, mp1])))
        a1 = value_of_area(area1)
        #   [1]
        mp2 = missing_area_point_for_three_points(np.array([maxXGivenY, pointY, point]))
        area2 = trim_area(area, corners_to_area(np.array([maxXGivenY, pointY, point, mp2])))

        a2 = value_of_area(area2)

        if a1 > a2: return area1
        return area2

    elif areaWanted == "xmax":
        maxYGivenX = greatest_qualifying_fourway_point_at_point(pointX, unqualifiedAreas, area, id2)
        mp1 = missing_area_point_for_three_points(np.array([maxYGivenX, pointX, point]))
        area1 = trim_area(area, corners_to_area(np.array([maxYGivenX, pointX, point, mp1])))
        return area1

    elif areaWanted == "ymax":
        maxXGivenY = greatest_qualifying_fourway_point_at_point(pointY, unqualifiedAreas, area, id1)
        mp2 = missing_area_point_for_three_points(np.array([maxXGivenY, pointY, point]))
        area2 = trim_area(area, corners_to_area(np.array([maxXGivenY, pointY, point, mp2])))
        return area2
    raise ValueError("invalid wanted area arg.")

"""
iterate from .min x (point) in `incrementDirections` at in

incrementDirections: increment|axis
"""
def greatest_qualifying_fourway_point_at_point(point, unqualifiedAreas, area, incrementDirections):

    if incrementDirections["increment"] >= 0:
        term = lambda s: True if point >= area[1,incrementDirections["axis"]] else False
    else:
        ##assert not (increment < 0), "invalid increment"
        term = lambda s: True if point <= area[1,incrementDirections["axis"]] else False

    inc = lambda p: (p[0], p[incrementDirections["axis"]] + incrementDirections["increment"]) if incrementDirections["axis"] == 1 else\
                (p[incrementDirections["axis"]] + incrementDirections["increment"], p[1])

    prevPoint = None
    point_ = (point[0], point[1])
    while point_in_area(point_, area):
        # check for qual
        if not is_point_qualified(unqualifiedAreas, point_): break
        prevPoint = point_
        point_ = inc(point_)
    return prevPoint

"""
description:
- given a 4 x 2 matrix with each element a corner, determines the lower left and upper
  right points.

arguments:
- fourSizedArea: 4 x 2 np.ndarray

return:
- 2 x 2 np.ndarray
"""
def corners_to_area(fourSizedArea):
    assert not (len(fourSizedArea.shape) != 2 or fourSizedArea.shape[0] != 4\
        or fourSizedArea.shape[1] != 2)

    # get min x, min y
    minXIndices, minYIndices = np.where(fourSizedArea[:,0] == np.min(fourSizedArea[:,0])),\
                            np.where(fourSizedArea[:,1] == np.min(fourSizedArea[:,1]))

    index = np.intersect1d(minXIndices, minYIndices)
    if len(index) != 1:
        print("MI")
        print(minXIndices)
        print(minYIndices)
        print("\nFSA")
        print(fourSizedArea)

        raise ValueError("points do not compose an area")
    minPoint = fourSizedArea[index[0]]

    #   frequency count, minPoint[0] occurs twice, minPoint[1] occurs twice
    locs = np.where(fourSizedArea[:,0] == minPoint[0])[0]
    if len(locs) != 2: raise ValueError("x coord. violation 1")

    locs = np.where(fourSizedArea[:,1] == minPoint[1])[0]
    if len(locs) != 2: raise ValueError("y coord. violation 1")

    # get max x, max y
    maxXIndices, maxYIndices = np.where(fourSizedArea[:,0] == np.max(fourSizedArea[:,0])),\
                            np.where(fourSizedArea[:,1] == np.max(fourSizedArea[:,1]))
    index = np.intersect1d(maxXIndices, maxYIndices)
    if len(index) != 1: raise ValueError("points do not compose an area")
    maxPoint = fourSizedArea[index[0]]

    #   frequency count, maxPoint[0] occurs twice, maxPoint[1] occurs twice
    locs = np.where(fourSizedArea[:,0] == maxPoint[0])[0]
    if len(locs) != 2: raise ValueError("x coord. violation 1")

    locs = np.where(fourSizedArea[:,1] == maxPoint[1])[0]
    if len(locs) != 2: raise ValueError("y coord. violation 1")
    return np.array([minPoint, maxPoint])

"""
"""
def trim_area(totalArea, area):
    assert is_valid_area(totalArea), "invalid total area"
    assert is_valid_area(area), "invalid area"
    # check min x
        # case: less than min x total
    if area[0,0] < totalArea[0,0]:
        area[0,0] = totalArea[0,0]

        # case: greater than max x total
    elif area[0,0] > totalArea[1,0]:
        area[0,0] = totalArea[1,0]

    # check min y
        # case: less than min y total
    if area[0,1] < totalArea[0,1]:
        area[0,1] = totalArea[0,1]

        # case: greater than max y total
    elif area[0,1] > totalArea[1,1]:
        area[0,1] = totalArea[1,1]

    # check max x
        # case: greater than max x total
    if area[1,0] > totalArea[1,0]:
        area[1,0] = totalArea[1,0]

        # case: less than min x total
    elif area[1,0] < totalArea[0,0]:
        area[1,0] = totalArea[0,0]

    # check max y
        # case: greater than max y total
    if area[1,1] > totalArea[1,1]:
        area[1,1] = totalArea[1,1]

        # case: less than min y total
    elif area[1,1] < totalArea[0,1]:
        area[1,1] = totalArea[0,1]
    return area

# TODO: matrix index accessor (using list of int.)
# TODO: approximate decimals to 5 places.

# TODO: needs to be tested.

# TODO: does not perform argument check on three points!
"""
description:
- determines the missing rectangular area point given three points.

arguments:
- threePoints: numpy array.

return:
- (float,float)
"""
def missing_area_point_for_three_points(threePoints):
    assert is_2dmatrix(threePoints)

    # find the center point
    index = center_point(threePoints)
    if index == -1: raise ValueError("[0] points given is invalid area")
    center = threePoints[index]

    # find the extreme-y point
    index2 = np.where(threePoints[:,0] == center[0])[0]
    if len(index2) != 2: raise ValueError("[1] points given is invalid area")
    yPoint = threePoints[index2[0] if index2[0] != index else index2[1]]

    # find the extreme-x point
    index3 = np.where(threePoints[:,1] == center[1])[0]
    if len(index3) != 2: raise ValueError("[2] points given is invalid area")
    xPoint = threePoints[index3[0] if index3[0] != index else index3[1]]

    return (xPoint[0], yPoint[1])

# TODO: test this 
def diagonal_type__2d(twoPoints): 
    """
    :return: 0 for not diagonal, 1 for left-hand diagonal, 2 for right-hand diagonal
    :rtype: int
    """
    assert is_2dmatrix(twoPoints), "invalid points"
    assert twoPoints.shape[0] == 2 and twoPoints.shape[1] == 2, "invalid shape for points"

    dx = round(twoPoints[0,0] - twoPoints[1,0],5)
    if dx == 0.0: return 0 
    dy = round(twoPoints[0,1] - twoPoints[1,1],5)
    if dy == 0.0: return 0 

    p1,p2 = None,None
    if twoPoints[0,0] < twoPoints[1,0]: 
        p1 = twoPoints[0]
        p2 = twoPoints[1] 
    else: 
        p1 = twoPoints[1]
        p2 = twoPoints[0]  

    if p1[1] > p2[1]: 
        return 1 
    return 2 

# TODO: untested, wrong
def other_points_for_two_points_in_area(twoPoints):
    assert is_2dmatrix(twoPoints), "invalid points"
    assert twoPoints.shape[0] == 2 and twoPoints.shape[1] == 2, "invalid shape for points"

    dt = diagonal_type__2d(twoPoints) 
    if dt == 0: 
        return None,None 

    px0,py0,px1,py1 = None,None,None,None 
    if dt == 1: 
        px0 = min(twoPoints[:,0])
        py0 = min(twoPoints[:,1])
        px1 = max(twoPoints[:,0])
        py1 = max(twoPoints[:,1])
    else: 
        px0 = min(twoPoints[:,0])
        py0 = max(twoPoints[:,1])
        px1 = max(twoPoints[:,0])
        py1 = min(twoPoints[:,1])        

    M = np.array([[px0,py0],[px1,py1]])
    return M 

# TODO: untested
def two_points_to_area(twoPoints):

    assert abs(twoPoints[0,0] - twoPoints[1,0]) > 10 ** -5, "two points cannot have same x-coord."
    assert abs(twoPoints[0,1] - twoPoints[1,1]) > 10 ** -5, "two points cannot have same y-coord."

    p0 = (max(twoPoints[:,0]), max(twoPoints[:,1]))
    p1 = (min(twoPoints[:,0]), max(twoPoints[:,1]))
    p2 = (min(twoPoints[:,0]), min(twoPoints[:,1]))
    p3 = (max(twoPoints[:,0]), min(twoPoints[:,1]))
    data = np.array([p0,p1,p2,p3])
    return corners_to_area(data)

def diagonal_line_from_corner_point_in_area(area, cornerPoint):
    q = quadrant_of_corner_point_in_area(area,cornerPoint)
    q2 = (q + 2) % 4
    return Line(np.array([q,q2]))

"""
description:
- given three points that form a right angle, determines the point in the middle.
  If the three points do not form a right angle, then returns -1 .
"""
def center_point(threePoints:np.ndarray, index = 0):
    if index < 0: raise ValueError("invalid index")
    if index >= threePoints.shape[0]: return -1

    indices = np.where(threePoints[:,0] == threePoints[index,0])[0] # TODO: check that equality works
    indices2 = np.where(threePoints[:,1] == threePoints[index,1])[0]
    if (True if len(indices) == 2 and len(indices2) == 2 else False):
        return index
    return center_point(threePoints, index + 1)

def area_to_corners(area):
    assert is_valid_area(area), "invalid area {}".format(area)
    upperLeft = [area[0,0], area[1,1]]
    lowerRight = [area[1,0], area[0,1]]
    return np.round(np.array([area[1], upperLeft, area[0], lowerRight]),5)

############################################ START: qualifying area search

"""
return:
- (start)::float,(end)::float,(distance)::float
"""
###
def largest_subrange_of_coincidence_between_ranges(r1,r2, roundDepth = 5):
    assert len(r1) == 2 and len(r2) == 2, "invalid ranges {} and {}".format(r1,r2)
    r1,r2 = sorted(r1), sorted(r2)

    ##if r2[0] < r1[0]: r1,r2 = r2,r1

    # use r2 as reference

    # case: r2[0] in r1
    if r2[0] >= r1[0] and r2[0] <= r1[1]:
        # check r2[1] in r1
        if r2[1] >= r1[0] and r2[1] <= r1[1]:
            rx = list(r2) + [r2[1] - r2[0]]
            return tuple(np.round(rx,roundDepth))
        else:
            rx = [r2[0],r1[1]] + [r1[1] - r2[0]]
            return tuple(np.round(rx,5))

    elif r2[1] >= r1[0] and r2[1] <= r1[1]:
        rx = [r1[0],r2[1]] + [r2[1] - r1[0]]
        return tuple(np.round(rx,5))
    return None,None,0.0

def is_proper_color(color):
    if color == None or type(color) in [int,float]: return False
    if len(color) != 3: return False #, "invalid length {} for color".format(len(color))
    for c in color:
        if type(c) not in [int,float]: return False #, "invalid type {} for color".format(type(c))
    return True

def area_to_pygame_rect_format(area):
    assert is_valid_area(area), "area {} is not valid!".format(area)
    lt = area[0]
    w,h = area[1,0] - area[0,0], area[1,1] - area[0,1]
    return np.array([lt, [w,h]])

## TODO:
"""
def area_from_point(p, areaWidth, areaHeight):
    q2 = (p[0] + areaWidth, p[1] + areaHeight)
    return np.array([p,q2])
"""

def point_to_area(p, areaWidth, areaHeight):
    assert is_valid_point(p), "invalid point {}".format(p)
    assert areaWidth > 0.0 and areaHeight > 0.0, "invalid width and height"
    w,h = areaWidth / 2.0, areaHeight / 2.0
    start = (p[0] - w, p[1] - h)
    end = (p[0] + w, p[1] + h)
    return np.array([start,end])



################################## START: cast functions

def float_func(s):
    q = float(s)
    return np.round(q, 5)

"""
always floor
"""
def int_func(s):
    return int(floorceil_to_n_places(float(s), 'f', places = 5))

def floorceil_to_n_places(f, mode, places = 5):
    assert places >= 0 and type(places) is int, "invalid places {}".format(places)
    assert mode in ["f","c"], "invalid mode {}".format(mode)

    d = f % 1.0
    w = f // 1.0
    q = 0.0
    lastDelta = 0.0
    for i in range(places):
        ten = 10 ** (-(i + 1))
        newD = d % ten
        q += (d - newD)
        if (d - newD) > 0: lastDelta = ten
        d = newD
    o = w + q
    if mode == "f": return o
    if f - o > 0:
        return o + lastDelta

################################## END: cast functions

#
cr = lambda x: round(float(x),5)

# TODO: rename to `iterable`
def vector_to_string(v, castFunc = int):
    assert castFunc in [int,float, float_func,cr], "invalid cast func"
    if len(v) == 0: return ""

    s = ""
    for v_ in v:
        s += str(castFunc(v_)) + ","
    return s[:-1]

# TODO: untested
def string_to_vector(s, castFunc = int):
    assert castFunc in [float, float_func, int,cr], "invalid cast func"

    def next_item(s):
        indo = len(s)
        for (i, q) in enumerate(s):
            if q == ",":
                indo = i
                break
        return s[:indo], s[indo + 1:]

    q = []
    while s != "":
        s1,s2 = next_item(s)
        v = castFunc(s1)
        q.append(v)
        s = s2
    return np.array(q)

def float_to_string(f,exclude_int:bool=False,exclude_exp:bool=False):
    sf = str(f)

    if exclude_int:
        q = sf.find(".")
        if q != -1:
            sf = sf[q+1:] 
    
    if exclude_exp:
        q = sf.find("E")
        if q != -1:
            sf = sf[:q] 
    return sf  

def write_vector_sequence_to_file(f,s,mode='w'):
    with open(f,mode) as fo:
        s2 = [vector_to_string(s_,cr) + "\n" for s_ in s]
        fo.writelines(s2)

######## start: some methods on bounds

euclidean_point_distance = lambda p1, p2: np.sqrt(np.sum((p1 - p2)**2))

'''
'''
def intersection_of_bounds(b1,b2):
    # check for each column
    assert is_2dmatrix(b1) and is_2dmatrix(b2),"invalid bounds {}\n\n\t{}".format(b1,b2)
    assert b1.shape == b2.shape, "must have equal shape"

    # check for each index
    bx = []
    for i in range(b1.shape[0]):
        q1,q2 = b1[i],b2[i]
        start,end,dist = largest_subrange_of_coincidence_between_ranges(q1,q2,5)
        if dist != 0:
            bx.append([start,end])
        else:
            return None
    return np.array(bx)

'''
b2 in b1?
'''
def bounds_is_subbounds(b1,b2):
    assert is_proper_bounds_vector(b1) and is_proper_bounds_vector(b2), "invalid args."
    assert b1.shape[0] == b2.shape[0], "invalid args. 2"

    def x(i):
        return b2[i,0] >= b1[i,0] and b2[i,1] <= b1[i,1]
    for i in range(b1.shape[0]):
        if not x(i): return False
    return True

"""
Calculates a subbound of bound `b` w/ start equal to the start of `b`,

boundRange := list({0,1})
"""
def subbound_of_bound(b, boundRange):
    assert is_2dmatrix(b), "invalid type for bounds"
    assert b.shape[1] == 2, "invalid shape for bounds"

    assert is_valid_point(boundRange), "[0] invalid bound range"
    assert np.all(np.array(boundRange) >= 0) and np.all(np.array(boundRange) <= 1), "[1] invalid bound range"
    assert len(boundRange) == 2, "[2] invalid bound range"

    diff = b[:,1] - b[:,0]

    s = b[:,0] + (diff * boundRange[0])
    e = b[:,0] + (diff * boundRange[1])
    return np.vstack((s,e)).T

def n_partition_for_bound(b, partition):

    q = (b[:,1] - b[:,0]) / float(partition)
    x0 = np.copy(b[:,0])
    partition_ = [x0]
    for i in range(partition):
        x1 = x0 + q
        x0 = np.copy(x1)
        partition_.append(x0)
    return np.array(partition_)

def invert_bounds(b):
    assert is_bounds_vector(b), "invalid bounds vector"
    b2 = np.empty((b.shape[0],2))

    b2[:,0] = np.copy(b[:,1])
    b2[:,1] = np.copy(b[:,0])
    return b2

def to_proper_bounds_vector(b):
    assert is_bounds_vector(b), "invalid bounds vector"

    b2 = np.empty((b.shape[0],2))
    b2[:,0] = np.min(b,axis=1)
    b2[:,1] = np.max(b,axis=1)
    return b2

def is_proper_bounds_vector(b):
    assert is_bounds_vector(b), "invalid bounds vector {}".format(b)
    return np.all(b[:,0] <= b[:,1])

def partial_invert_bounds(b, indices):
    return -1

# TODO: untested
def extremum_for_points(a):
    '''
    calculates the `n` extreme (min & max) values for the
    `m x n` matrix `a`.
    '''

    assert is_2dmatrix(a), "invalid points"
    mi = np.min(a,axis = 0)
    ma = np.max(a,axis = 0)
    b = np.empty((a.shape[1],2))
    b[:,0] = mi
    b[:,1] = ma
    return b

def point_in_bounds(b,p):
    '''
    determines if point `p` in bounds `b`

    :param b: proper bounds vector
    :param p: vector
    :rtype: bool
    '''

    assert is_proper_bounds_vector(b), "invalid bounds vector"
    assert is_vector(p), "invalid point"
    assert b.shape[0] == p.shape[0]
    return np.all(b[:,0] <= p) and np.all(b[:,1] >= p)

def point_in_bounds_(b,p):
    if b.shape[0] == 1:
        return np.all(b[0,0] <= p) and np.all(b[0,1] >= p)
    return point_in_bounds(b,p)

def point_in_improper_bounds(parentBounds,bounds,p):

    if is_proper_bounds_vector(bounds):
        assert bounds_is_subbounds(parentBounds,bounds), "invalid parent bounds"
        return point_in_bounds(bounds,p)

    ibs = split_improper_bound(parentBounds,bounds,checkPoints = True)

    rf = lambda f,i: (f >= ibs[0][i,0] and f <= ibs[0][i,1]) or\
                (f >= ibs[1][i,0] and f <= ibs[1][i,1])

    for (i,p_) in enumerate(p):
        if not rf(p_,i): return False
    return True


def point_difference_of_improper_bounds(improperBounds,parentBounds):
    """
    point difference of improper bounds is a non-negative vector
    with each i'th element the positive distance between the two points
    of the `improperBounds` in the context of the `parentBounds`. 
    
    :param improperBounds: bounds located in parentBounds
    :param parentBounds: bounds used as context for `improperBounds`. 
    """

    assert improperBounds.shape == parentBounds.shape, "bounds must have equal shape"
    if is_proper_bounds_vector(improperBounds):
        return improperBounds[:,1] - improperBounds[:,0]

    diff = []
    for i in range(improperBounds.shape[0]):
        r = improperBounds[i]
        j = parentBounds[i]

        # proper dim.
        if r[0] < r[1]:
            d = r[1] - r[0]
        # improper dim.
        else:
            d = j[1] - r[0]
            d += (r[1] - j[0])
        diff.append(d)
    return np.array(diff)


# TODO: unused
'''
'''
def split_improper_bound(properBounds,improperBounds,checkPoints = True):
    if checkPoints:
        assert point_in_bounds(properBounds,improperBounds[:,0]), "end0 of bounds not in proper bounds,\npoint {}\nbounds {}".format(improperBounds[:,0],properBounds)
        assert point_in_bounds(properBounds,improperBounds[:,1]), "end1 of bounds not in proper bounds"

    q1,q2 = [],[]
    for i in range(properBounds.shape[0]):
        r = improperBounds[i]
        # proper dim.
        if r[0] < r[1]:
            s1,s2 = np.copy(r),np.array([np.nan,np.nan])
        # improper dim.
        else:
            s1,s2 = np.array([r[0],properBounds[i,1]]), np.array([properBounds[i,0],r[1]])
        q1.append(s1)
        q2.append(s2)
    return np.array(q1),np.array(q2)

def point_on_bounds_by_ratio_vector(b,rv):
    x = rv * (b[:,1] - b[:,0])
    return b[:,0] + x

def point_on_improper_bounds_by_ratio_vector(parentBounds,bounds,rv,roundDepth = 5):

    if is_proper_bounds_vector(bounds):
        return point_on_bounds_by_ratio_vector(bounds,rv)

    ib = split_improper_bound(parentBounds,bounds,checkPoints = True)
    q = point_difference_of_improper_bounds(bounds,parentBounds)
    pd1 = ib[0][:,1] - ib[0][:,0] # point diff
    s = rv * q # point add
    p2 = np.copy(bounds[:,0])

    for (i,s_) in enumerate(s):
        d = pd1[i] - s_

        # if addition is greater than first half
        if d < 0:
            px = parentBounds[i,0] - d
        else:
            px = p2[i] + s_
        p2[i] = px

    return np.round(p2,roundDepth)

def vector_ratio(bound,point):
    assert point_in_bounds(bound,point), "point not in bounds"
    return (point - bound[:,0]) / (bound[:,1] - bound[:,0])

'''
'''
def vector_ratio_improper(parentBounds,bounds,point):
    ib = split_improper_bound(parentBounds,bounds,checkPoints = True)
    assert len(ib) > 1, "bounds not improper"
    pd = point_difference_of_improper_bounds(bounds,parentBounds)

    def h(p_,i):
        t = pd[i]

        if p_ >= ib[0][i,0] and p_ <= ib[0][i,1]:
            x = p_ - ib[0][i,0]
        else:
            x = (ib[0][i,1] - ib[0][i,0]) + (p_ - parentBounds[i,0])
        return x / t

    v_ = []
    for (i,p_) in enumerate(point):
        v_.append(h(p_,i))
    return np.round(v_,5)

def bounded_point_to_hop_counts(parentBound,bound,p,h):
    '''
    maps each value of the point `p` relative to the bound to the number
    '''
    r = vector_ratio_improper(parentBound,bound,p)
    r = np.array([r_ if not np.isnan(r_) else 1. for r_ in r])
    h_ = 1. / h
    q = np.array([int(round(r_ / h_)) for r_ in r]) 
    return q

def refit_points_for_new_bounds(points,oldBounds,newBounds):
    q = []
    for p_ in points:
        q.append(refit_point_for_new_bounds(p_,oldBounds,newBounds))
    return np.array(q)

def refit_point_for_new_bounds(p,oldBounds,newBounds):
    vr = vector_ratio(oldBounds,p)
    return point_on_bounds_by_ratio_vector(newBounds,vr)

def area_of_bounds(bounds):
    assert is_proper_bounds_vector(bounds), "not proper bounds vector"
    pd = bounds[:,1] - bounds[:,0]
    return np.product(pd)

######## end: some methods on bounds



from morebs2 import g2tdecomp
from .graph_basics_test_cases import * 
import unittest

'''
python -m morebs2.tests.g2tdecomp_test  
'''
class TNodeClass(unittest.TestCase):

    def test__TNode__collate_keys_case1(self):
        D = graph_case_12() 
        gd = g2tdecomp.G2TDecomp(D)
        gd.decompose()

        prg = g2tdecomp.prg__constant(0) 
        prg2 = g2tdecomp.prg__n_ary_alternator(start=0) 
        prg3 = g2tdecomp.prg__n_ary_alternator(start=1) 
        prg4 = g2tdecomp.prg__n_ary_alternator(s0=3,s1=25,start=1) 

        tn = gd.decompositions[0]
        qx = g2tdecomp.TNode.collate_keys(tn,True,None)
        qx2 = g2tdecomp.TNode.collate_keys(tn,False,None)
        qx3 = g2tdecomp.TNode.collate_keys(tn,False,prg)
        qx4 = g2tdecomp.TNode.collate_keys(tn,False,prg2)
        qx5 = g2tdecomp.TNode.collate_keys(tn,False,prg3)
        qx6 = g2tdecomp.TNode.collate_keys(tn,True,prg4)

        assert qx == [1, 0, 2, 8, 3, 4, 6, 7, 3, 9, 10, 10, 10, 9, 4]
        assert qx6 == [1, 2, 8, 0, 7, 6, 9, 3, 4, 3, 10, 10, 10, 4, 9]
        assert qx2 == [1, 1, 0, 3, 10, 9, 10, 3, 0, 4, 10, 4, 0, 1, 2,\
            6, 2, 7, 2, 1, 8, 3, 8, 9, 10, 4, 10, 9, 8, 1]
        assert qx3 == [1, 0, 3, 10, 9, 4, 2, 6, 7, 8]

        assert qx4 == [1, 0, 3, 10, 9, 10, 0, 4, 4, 1, 2, 6, 7, 2, 8, 3, 9, 4, 9, 1]
        assert qx5 == [1, 1, 0, 3, 10, 9, 3, 4, 10, 0, 2, 6, 2, 7, 1, 8, 8, 10, 10, 8]
        return 


class G2TDecompClass(unittest.TestCase):

    def test__G2TDecomp__decomp_case1(self):
        # soln 
        a01 = defaultdict(list, {0: [0, 2], 1: [0, 2], 3: [0, 0], 4: [0, 0], 2: [0, 2]})
        a0 = (6,a01)

        a11 = defaultdict(list, {1: [0, 2], 0: [0, 2], 3: [0, 0], 4: [0, 0], 2: [0, 0]})
        a1 = (4,a11)

        a21 = defaultdict(list, {2: [0, 2], 0: [0, 0], 1: [0, 0]})
        a2 = (2,a21) 

        a31 = defaultdict(list, {3: [0, 2], 0: [0, 0], 1: [0, 0]})
        a3 = (2,a31)

        a41 = defaultdict(list, {4: [0, 2], 0: [0, 0], 1: [0, 0]})
        a4 = (2,a41)

        sol0 = {0:a0,1:a1,2:a2,3:a3,4:a4} 

        D = graph_case_15() 
        gd = g2tdecomp.G2TDecomp(D,child_capacity=2)
        gd.decompose()

        ks = set()
        for x in gd.decompositions: 
            dx, md,_ = g2tdecomp.TNode.dfs(x,display=False)
            q = sol0[x.idn] 
            ec = g2tdecomp.edge_count(dx)
            ncm = g2tdecomp.nc_degree_map(dx) 
            assert ec == q[0] 
            assert ncm == q[1]
            ks |= {x.idn}
        
        assert ks == set(sol0.keys())

    def test__G2TDecomp__decomp_case2(self):

        # sol'n 
        r = defaultdict(list, \
            {0: [0, 4], 1: [0, 2], 5: [0, 0], \
            6: [0, 0], 2: [0, 0], 3: [0, 0], 4: [0, 0]})
        sol0 = (6,r)

        r7 = defaultdict(list, \
            {7: [0, 2], 12: [0, 2], 38: [0, 1], \
            11: [0, 1], 31: [0, 0], 41: [0, 0], \
            20: [0, 2], 15: [0, 0], 64: [0, 0]})

        sol7 = (8,r7)

        D = graph_case_11() 
        gd = g2tdecomp.G2TDecomp(D) 
        gd.decompose()

        for x in gd.decompositions: 
            dx, md,_ = g2tdecomp.TNode.dfs(x,display=False)

            if x.idn == 0: 
                q1 = g2tdecomp.edge_count(dx) 
                q2 = g2tdecomp.nc_degree_map(dx) 
                assert q1 == sol0[0]
                assert q2 == sol0[1]

                sz0 = g2tdecomp.TNode.size_count(x)
                assert sz0 == 7 

            elif x.idn == 7: 
                q1 = g2tdecomp.edge_count(dx) 
                q2 = g2tdecomp.nc_degree_map(dx) 
                assert q1 == sol7[0]
                assert q2 == sol7[1]

                sz7 = g2tdecomp.TNode.size_count(x)
                assert sz7 == 9 

            elif x.idn == 11: 
                q1 = g2tdecomp.edge_count(dx) 
                q2 = g2tdecomp.nc_degree_map(dx) 
                assert q1 == 1

                sz11 = g2tdecomp.TNode.size_count(x)
                assert sz11 == 2 

        assert len(gd.decompositions) == 16, "actual {}".format(len(gd.decompositions))

    def test__G2TDecomp__decomp_case3(self):
        """
        PRG case 
        """

        sol5 = (12,defaultdict(set, \
            {5: {0}, 0: {1, 2}, 2: {6}, \
            6: {7}, 7: set(), 1: {8}, 8: {9, 3}, \
            9: {10}, 10: {4}, 4: set(), 3: {10}}))

        sol2 = (6,defaultdict(set, \
            {2: {1, 7}, 1: {0}, 0: {3}, \
            3: set(), 7: {6}, 6: set()}))

        sol1 = (8,defaultdict(set, \
            {1: {2}, 2: {0}, 0: {4}, 4: {10}, \
            10: {9, 3}, 9: {8}, 8: set(), 3: set()}))

        sol9 = (1,defaultdict(set, {9: set()}))

        sols = {0:sol5,1:sol2,2:sol1,3:sol9}

        D = graph_case_12() 
        prg4 = g2tdecomp.prg__n_ary_alternator(s0=3,s1=25,start=1) 

        gd = g2tdecomp.G2TDecomp(D,prg=prg4)
        gd.decompose()

        for i in range(4): 
            x = gd.decompositions[i]
            sz = g2tdecomp.TNode.size_count(x) 
            dx, md,_ = g2tdecomp.TNode.dfs(x,display=False)
            assert sols[i][0] == sz, "wrong at {}".format(i)
            assert sols[i][1] == dx 
        return  


if __name__ == '__main__':
    unittest.main()


from .matrix_methods import *

########## START: stat measures

def standard_variance_on_sequence(npSeq):
    assert len(npSeq.shape) == 1, "invalid dim., want 1"
    return npSeq.var()

# TODO: in testing
def opposing_deviation_on_sequence(npSeq):
    assert len(npSeq.shape) == 1, "invalid dim., want 1"
    m = npSeq.mean()
    diff = np.array([x - m for x in npSeq])
    return sum(diff)

# TODO: opposing variance w/in e-ball

########## END: stat measures

########## START: point measures

zero_div = lambda num, denum, default: default if denum == 0 else num / denum

point_distance = lambda p1, p2: math.sqrt((p1[0] - p2[0]) **2 + (p1[1] - p2[1]) ** 2)
'''
standard point-distance formula

:param p1: two-dimensional point
:param p2: two-dimensional point
'''

point_distance_one_to_rest = lambda p1, ps: sum(point_distance(p1, ps_) for ps_ in ps)
'''
summation of total distances from one point to iterable of points

:param p1: 2-d iterable
:param ps: iterable<2-d points>
'''

"""
description:
- determines which points in `ps` are <= `threshold` distance from .`p1`
"""
point_distance_below_threshold_one_to_rest = lambda p1, threshold, ps: np.array([ps_ for ps_ in ps if point_distance(p1, ps_) <= threshold])

# TODO: check for 0 case
point_slope = lambda x, x2: (x2[1] - x[1]) / (x2[0] - x[0]) if  abs(x2[0] - x[0]) >= 10 ** -5 else 0

midpoint = lambda x, x2: np.array([(x[0] + x2[0]) / 2, (x[1] + x2[1]) / 2])

def angle_between_two_points(startX, endX):
    if equal_iterables(startX, endX):
        return 0

    l = point_distance(startX, endX)
    opposite = endX[1] - startX[1]

    q = opposite / float(l)
    return math.degrees(np.arcsin(q))

"""
"""
def angle_between_two_points_clockwise(startX, endX):
    angle = angle_between_two_points(startX, endX)
    if angle == 0:
        if endX[0] <startX[0]: return 180.0
        return 0.0

    # case: forward
    if startX[0] < endX[0]:
        if angle > 0:
            return angle
        else:
            return angle % 360
    else:
        return 180 - angle
    return angle

def angle_between_two_angles(angle, ref, end):
    assert angle >= 0 and angle < 360, "invalid angle"
    assert ref >= 0 and ref < 360, "invalid angle"
    assert end >= 0 and end < 360, "invalid angle"

    # case 0
    if ref > end:
        stat = angle >= ref and angle < 360 # ref to 360
        stat2 = angle >= 0 and angle <= end # 0 to end
        if stat or stat2: return True
        return False

    # case 1
    else:
        return True if angle >= ref and angle <= end else False

def positive_complement_of_angle(angle):
    assert angle >= 0 and angle < 360, "invalid angle"
    return 360 - angle

def quadrant_of_degree(degree):
    degree = degree % 360.0

    if degree <= 90:
        return 0

    if degree <= 180:
        return 1

    if degree <= 270:
        return 2

    return 3

def contiguous_point_slopes(points):
    '''
    Calculates the slope b/t every contiguous point pairs.

    :param points: sequence of points
    :type points: np.ndarray
    '''
    slopes = []
    for i in range(0, points.shape[0] - 1):
        p1, p2 = points[i], points[i + 1]
        slopes.append(point_slope(p1, p2))
    return np.array(slopes)

# TODO: refactor for n-degree??
def contiguous_point_acceleration(points):
    s1 = contiguous_point_slopes(points)
    return contiguous_point_slopes(s1)

########## END: point measures

########## START: array measures

def partition_average_of_array(a2d, targetAxis, ph:int):
    if len(a2d.shape) != 2:
        raise ValueError("array must be 2-d")
    if targetAxis != 0 and targetAxis != 1:
        raise ValueError("target axis is invalid, 0 OR 1")

    average = []
    index = 0
    otherIndex = 0 if targetAxis == 0 else 1

    while index < a2d.shape[otherIndex]:
        nextIndex = index + ph

        if otherIndex == 0:
            mv = a2d[index:nextIndex].mean(axis = targetAxis)
        else:
            mv = a2d[:,index:nextIndex].mean(axis = targetAxis)

        average.append(mv)
        index = nextIndex
    return np.array(average)

# TODO: make noise to point
def visualize_adding_different_noise_amplitude_to_line(line, folderPath, numImages, numAmplitudes = 10, maxAmplitude = 10, signRestriction = 0):
    assert numImages >= numAmplitudes and  numImages // numAmplitudes == numImages / numAmplitudes, "invalid # of images and amplitudes!"

    def plot_and_save(filePath):
        return -1

    numPicsPer = numImages // numAmplitudes
    amplitudeIncrement = maxAmplitude / numAmplitudes

    line.form_point_sequence("exact")

    for i in range(numAmplitudes):
        q = amplitudeIncrement * (i + 1)
        for j in range(numPicsPer):
            return -1
        return -1
    return -1

########## END: point measures

########## START: polynomial density measures

from collections import Counter

# TODO: draft
def polynomial_density_measure_1(points):
    slopes = list(contiguous_point_slopes(points))

    l = points.shape[0] - 1
    q = Counter(slopes)
    q = normalize_dict_values_by_function(q, lambda x: zero_div(x,l,1))
    # TODO: add noise modification here
    vq = [v for v in q.values()]
    ###
    return len(q) / l

def normalize_dict_values_by_function(d, f):
    for k,v in d.items():
        d[k] = f(v)
    return d

########## END: polynomial density measures

########## START: formulas for equations and geometric shapes
def quadratic_formula_solve_for_reals(a,b,c):
    """
    quadratic formula given a,b,c

    :return: sol. set,sol. exist
    :rtype: (list,bool)
    """
    fac = 4 * a * c
    if b ** 2 < fac:
        return [], False

    q = math.sqrt(b ** 2 - fac)
    x1,x2 = (-b + q) / (2 * a), (-b - q) / (2 * a)
    return [x1,x2], True

# TODO: not tested.
#       easier approach is distance(p, centerPoint) < radius
def is_point_in_circle(p, centerPoint, radius):

    if not (p[0] >= centerPoint[0] - radius and p[0] <= centerPoint[0] + radius):
        return False

    a = 1
    b = -2 * centerPoint[1]
    c = p[0] ** 2 + centerPoint[0] ** 2 - 2 * centerPoint[0] * p[0] - radius ** 2 + centerPoint[1] ** 2

    soln, stat = quadratic_formula_solve_for_reals(a,b,c)
    if not stat: return stat

    if p[1] >= min(soln) and p[1] <= max(soln): return True
    return False

def is_line_in_circle(line, circleCenter, circleRadius):

    # endpoint [0]
    if is_point_in_circle(line.endpoints[0], circleCenter, circleRadius):
        return True

    # endpoint [1]
    if is_point_in_circle(line.endpoints[1], circleCenter, circleRadius):
        return True

    return False

########## END: formulas for equations and geometric shapes

def euclidean_point_distance_of_bounds(parentBounds,bounds):

    if is_proper_bounds_vector(bounds):
        ed = euclidean_point_distance(bounds[:,1],bounds[:,0])
    else:
        pd = point_difference_of_improper_bounds(bounds, parentBounds)
        z = np.zeros((bounds.shape[0],))
        ed = euclidean_point_distance(z,pd)
    return ed



from collections import defaultdict 
from copy import deepcopy 
from .globalls import invert_map__seqvalue 

def is_undirected_graph(d):
    assert type(d) in {defaultdict,dict}
    
    for k,v in d.items(): 
        for v_ in v: 
            if k not in d[v_]: return False 
    return True

def is_directed_graph(d): 
    return not is_undirected_graph(d)

def graph_childkey_fillin(d): 
    kx = list(d.keys()) 
    for k in kx: 
        V = d[k]
        for v in V: 
            if v not in d: 
                d[v] = set()

def edge_count(d): 
    return sum([len(v) for v in d.values()])

def index_of_element_in_setseq(s,n):
    for (i,s_) in enumerate(s): 
        if n in s_: return i 
    return None 

def directed_edge_partition(d,k,V): 
    partition = [set(),set()]
    for v_ in V: 
        if k not in d[v_]: 
            partition[1] |= {v_} 
        else: 
            partition[0] |= {v_} 
    return partition

"""
neighbor-child degree map. A node n's neighbor-child degree is 
    [|neighbors(n)|,|children(n)|]

:return: map of node idn to its neighbor-child degree
:rtype: defaultdict(list) 
"""
# TODO: test 
def nc_degree_map(G): 

    G_ = deepcopy(G) 
    d = defaultdict(list)
    for k in G_.keys(): 
        d[k] = [0,0]

    for k,v in G_.items():
        q = list(v)  
        for v_ in q: 
            # case: v_ is child 
            if k not in G_[v_]: 
                d[k][1] += 1 
            else: 
                d[k][0] += 1 
                d[v_][0] += 1 
            G_[v_] -= {k} 
        G_[k] = set() 
    return d

"""
neighbor-parent degree map. A node n's neighbor-parent degree is 
    [|neighbors(n)|,|parent(n)|]

:return: map of node idn to its neighbor-child degree
:rtype: defaultdict(list) 
"""
# TODO: test 
def np_degree_map(G): 

    G_ = deepcopy(G) 
    d = defaultdict(list)
    for k in G_.keys(): 
        d[k] = [0,0]

    for k,v in G_.items():
        q = list(v) 
        for v_ in q: 
            # case: neighbors 
            if k in G_[v_]: 
                d[k][0] += 1
                d[v_][0] += 1  
                G_[v_] -= {k}
            else: 
                d[v_][1] += 1 
        G_[k] = set() 
    return d 


def connected_to(G,n): 
    ks = list(G[n]) 
    for k,v in G.items():
        if k == n: continue 
        if n in v: ks.append(k) 
    return set(ks)

def parents_of(G,n): 
    p = [] 
    for k,v in G.items(): 
        if k == n: continue 
        if n in v and k not in G[n]: 
            p.append(k) 
    return set(p)

def children_of(G,n): 
    p = G[n]
    l = []
    for p_ in p: 
        if n not in G[p_]: 
            l.append(p_)
    return set(l) 

def doubly_connected(G,n): 
    ks = list(G[n]) 
    i = 0 
    while i < len(ks): 
        v = G[ks[i]] 
        if n not in v: 
            ks.pop(i)
        else: 
            i+=1 
    return set(ks)

def cpc_order(G,n): 
    return connected_to(G,n),parents_of(G,n),children_of(G,n) 

def delete_graph_edges(G,E,is_directed:bool=False): 
    
    while len(E) > 0: 
        e1 = E.pop() 
        G[e1[0]] = G[e1[0]] - {e1[1]} 

        if not is_directed: 
            e2 = (e1[1],e1[0]) 
            G[e2[0]] = G[e2[0]] - {e2[1]} 
            E = E - {e2}
    return

def flatten_setseq(s): 
    s_ = set() 
    for s2 in s: s_ |= s2 
    return s_

class GraphComponentDecomposition:

    def __init__(self,d): 
        """
        Calculates the components of a graph `d`. If `d` is a directed graph, 
        output (see class variable `components`) is a list of directed components. 
        Each directed component is a list of nodesets. The nodesets are ordered 
        in ascending order based on its depth rank. Additionally, the class variable 
        `cyclic_keys` maps each node to the nodeset that is its cyclic children. A cyclic 
        child is a parent of a node `n` but a child of another node `n2` connected to `n`. 
        The `cyclic_keys` map is used to determine if a directed component is of a 
        cyclic ordering, as opposed to static order. In a directed component `C` of 
        cyclic ordering, there exists at least a pair of nodes (`n1`,`n2`). The nodes 
        `n1` and `n2` belong to different nodesets of indices `i` and `i+j`. All nodes 
        in the nodesets of `C[i:i+j+1]` are of the same depth. If `d` is an undirected 
        graph, every element of `components` is a nodeset. If a directed component is 
        of static ordering, the depth rank of a node `n` in it is the index of the nodeset 
        that `n` is present in. 

        :param d: map representation of a graph 
        :type d: defaultdict(<set>)
        """

        assert type(d) == defaultdict
        assert d.default_factory == set 

        graph_childkey_fillin(d) 
        self.d = d 
        self.d_ = deepcopy(self.d)
        self.is_directed = is_directed_graph(self.d)

        self.components = []
        self.key_queue = [] 
        self.key_cache = set()

        self.cyclic_keys = defaultdict(set) 
        self.cyclic_keys_inverted = defaultdict(set) 

        self.finstat = False 

    def finstat_(self): 
        self.finstat = len(self.key_cache) == len(self.d) 
        return self.finstat

    '''
    main method 
    '''
    def decompose(self): 
        while not self.finstat:
            self.next_key()
        q = invert_map__seqvalue(self.cyclic_keys) 
        for k,v in q.items(): 
            self.cyclic_keys_inverted[k] = set(v) 
        return 

    def next_key(self): 
        if self.finstat_(): 
            return 

        if not self.is_directed: 
            self.undirected_next_key() 
            return 

        if len(self.key_queue) == 0: 
            x = set(self.d_.keys()) - self.key_cache
            x = sorted(list(x))
            self.init_decomp(x[0]) 
            return 

        k = self.key_queue.pop(0) 
        self.update_dircomp_by_kv_pair(k)

    def undirected_next_key(self): 

        if len(self.key_queue) == 0: 
            x = set(self.d_.keys()) - self.key_cache
            x = sorted(list(x))
            self.key_queue.append(x[0])

        q = self.key_queue.pop(0) 
        self.merge_by_edges(q)
        self.key_cache |= {q}  
        rem = self.d[q] - self.key_cache 
        self.key_queue.extend(list(rem)) 
        return 

    def depth_rank_map(self): 
        if not self.is_directed: return None 

        def update_map(Dx): 
            for k,v in Dx.items(): 
                D[k] += v 

        D = defaultdict(int) 
        
        for i in range(len(self.components)): 
            dx = self.depth_rank_map__component(i)
            update_map(dx)
        return D 

    def cyclic_component_indices(self): 
        if not self.is_directed: 
            return [] 

        j = [] 
        for i in range(len(self.components)): 
            if self.is_dcomponent_cyclic(i):
                j.append(i)
        return j 

    def depth_rank_map__component(self,ci): 
        D = defaultdict(int) 
        q = self.dcomponent_cyclic_indices(ci) 
        C = self.components[ci] 

        indexia = (-2,-2) if len(q) == 0 else q.pop(0)
        d = 0 
        for (i,c) in enumerate(C): 
            # case: nodeset in cycle, all of same depth 
            if i >= indexia[0] and i <= indexia[1]:
                for c_ in c: 
                    D[c_] = d
                continue 
            # case: move on to the next indexia 
            elif i == indexia[1] + 1: 
                indexia = (-2,-2) if len(q) == 0 else q.pop(0)
                d += 1
            for c_ in c: 
                D[c_] = d 
            d += 1 
        return D 

    def dcomponent_cyclic_indices(self,ci): 
        c = self.components[ci] 
        prev = 0 

        # each element is (head index,tail index)
        soln = []
        for (i,c_) in enumerate(c): 
            n_ = set()
            V = set() 
            for n in c_: 
                if n in self.cyclic_keys: 
                    n_ |= {n} 
                    V |= self.cyclic_keys[n] 
                    break 
            
            if len(n_) == 0: 
                continue 

            k = None 
            for j in range(prev,i): 
                if len(c[j].intersection(V)) > 0: 
                    k = j 
                    break 

            if type(k) == type(None):
                print("[!] weird") 
                continue 

            element = (j,i)
            soln.append(element) 
        return soln 
            
    def is_dcomponent_cyclic(self,ci): 
        if not self.is_directed: 
            return False 

        q = flatten_setseq(self.components[ci])
        for q_ in q: 
            ck = self.cyclic_keys[q_] 
            if ck.intersection(q): 
                return True  
        return False 
    
    #---------------- undirected case 

    def merge_by_edges(self,k): 
        i = index_of_element_in_setseq(self.components,k)
        kx = {k} | self.d[k]

        if type(i) == type(None): 
            self.components.append(kx)
            return 
        self.components[i] |= kx 

    #------------------------------ directed update methods 

    def init_decomp(self,k): 
        self.key_cache |= {k}

        partition = directed_edge_partition(self.d_,k,list(self.d_[k]))

        if len(partition[0]) + len(partition[1]) == 0: 
            self.components.append([{k}])
            return 

        nc = self.doubly_connected_subsets(partition[1])
        if len(nc) == 0: 
            self.components.append([{k} | partition[0]]) 
        else: 
            nc = [[{k} | partition[0],nc_] for nc_ in nc] 
            self.components.extend(nc) 

        self.queue_update(partition)
        self.graph_edge_update(k,partition)

    def update_dircomp_by_kv_pair(self,k): 
        V = [k] + list(self.d_[k]) 

        partition = directed_edge_partition(self.d_,V[0],V[1:]) 

        partition[0] |= {k}
    
        i = 0 
        new_comps = [] 
        cc = set()
        while i < len(self.components): 
            o1,new_comp,cyclic_children = self.update_dircomp(i,partition)
            if type(o1) == type(None): 
                i += 1 
                continue 
            self.components.pop(i)
            new_comps.extend(new_comp)
            cc |= cyclic_children
        self.components.extend(new_comps) 
        partition[0] -= {k}
        self.key_cache |= {k} 
        self.queue_update(partition)
        self.graph_edge_update(k,partition)
        self.update_cyclic_children(k,cc)

    def update_cyclic_children(self,k,cc):
        if len(cc) == 0: return  
        self.cyclic_keys[k] |=  cc 

    def update_dircomp(self,i,partition): 

        o1,o2,o3 = self.check_dircomp_with_partition(i,partition)
        
        # case: no connection 
        if type(o1) == type(None):
            return o1,[],o2 

        # case: connection, reform 
            # subcase: there are cyclic children 
        cp = deepcopy(self.components[i]) 
        cp[o1] = cp[o1] | partition[0] 

        if len(o3) == 0: 
            return o1,[cp],o2 

        new_components = self.new_components_from_children_(cp,o1,o3) 
        return o1,new_components,o2  

    def new_components_from_children_(self,cp,i,c_q): 
        
        nc = []
        # case: neighbor set in cp is last element 
        if i == len(cp) - 1: 
            for x in c_q: 
                cp2 = deepcopy(cp) 
                cp2.append(x) 
                nc.append(cp2)
            return nc         

        # case: not of the last 
        ns1 = flatten_setseq(cp[i+1:])
        nc = [deepcopy(cp)] 
        for x in c_q: 
            chk = x - ns1 
            # component has unrelated children from previous 
            # make new component without cp[i+1:]
            if len(chk) > 0:     
                cp2 = deepcopy(cp[:i+1]) 
                cp2.append(x) 
                nc.append(cp2) 
                continue 
        return nc 

    """
    return: 
    - index of intersected nodeset with partition neighbors in component
    - set of cyclic children 
    - remaining children (not cyclic) grouped into neighbors and lones 
    """
    def check_dircomp_with_partition(self,ci,partition): 

        C_d = self.components[ci]
        # check for intersection of neighbors with any in 
        # directed component 
        i = -1 
        for (j,c) in enumerate(C_d):
            cx = c.intersection(partition[0]) 
            if len(cx) > 0: 
                i = j 
                break 
        
        # case: no relation, nothing to do 
        if i == -1: return None,None,None 

        # check that there are no children before the i'th 
        # index 
        cyclic_children = set()
        for c in partition[1]: 
            j = index_of_element_in_setseq(C_d,c) 
            if type(j) == type(None): continue 

            if j < i: 
                cyclic_children |= {c}

        p2 = partition[1] - cyclic_children
        q = self.doubly_connected_subsets(p2)
        return i,cyclic_children,q

    def doubly_connected_subsets(self,ns): 
        s = [{ns_} for ns_ in ns]  
        for ns_ in ns: 
            q = index_of_element_in_setseq(s,ns_) 
            dconn = doubly_connected(self.d,ns_)

            to_remove = []
            for d in dconn: 
                q2 = index_of_element_in_setseq(s,d) 
                if type(q2) != type(None): 
                    s[q] = s[q] | s[q2] 
                    to_remove.append(q2) 
            
        return list(s)


    #---------------- indexing and travel memory

    def index_of(self,k): 
        if not self.is_directed: 
            return index_of_element_in_setseq(self.components,k)

        for (i,d_) in enumerate(self.components): 
            i2 = index_of_element_in_setseq(d_,k) 
            if type(i2) != type(None): 
                return (i,i2) 
        return None 

    def graph_edge_update(self,k,partition): 
        px0 = set([(k,p) for p in partition[0]])
        px1 = set([(k,p) for p in partition[1]])

        delete_graph_edges(self.d_,px0,is_directed=False)
        delete_graph_edges(self.d_,px1,is_directed=True)

    def queue_update(self,partition): 
        q = partition[0] | partition[1]

        for q_ in q: 
            if q_ not in self.key_cache: 
                self.key_queue.append(q_)


from morebs2 import numerical_generator
import numpy as np
from collections import OrderedDict

import unittest

'''
python -m morebs2.tests.numerical_generator_test  
'''
class TestNumericalGeneratorClass(unittest.TestCase):

    def test_CycleMap__random_map(self):
        vr = numerical_generator.CycleMap.random_cycle_map(5)
        ##print("VR")
        ##print(vr)

        f = numerical_generator.CycleMap.is_valid_map(vr)
        ##print("F: ", f)
        assert f, "random cycle map is not cycle"

        q = OrderedDict()
        x = [(3,1),(1,3),(2,4),(4,2)]
        for k,v in x:
            q[k] = v

        f2 = numerical_generator.CycleMap.is_valid_map(q)
        ##print("F2: ", f2)
        assert not f2, "non-cyclic map"
        return

    def test__generate_possible_binary_sequences(self):
        g = numerical_generator.generate_possible_binary_sequences(5, [])
        g = list(g)
        assert len(g) == 2 ** 5, "incorrect generation"

        # uncomment for viewing
        '''
        for g_ in g:
            print(g_)
        '''

    def test__random_npoint_from_point_in_bounds(self):

        bounds = np.array([[-100,100.0],\
                        [-100,100.0],\
                        [-100,100.0],\
                        [-100,100.0],\
                        [-100,100.0],\
                        [-100,100.0]])
        r = 200.0

        # case point on >= 1 of the bound extremes
        p1 = np.array([-100,100,-99,80,95.0,-100])
        for i in range(100):
            p2 = numerical_generator.random_npoint_from_point_in_bounds(bounds,p1,r)
            assert type(p2) != type(None)
        return

    def test__random_npoint_from_point(self):
        r = 200.0

        # case point on >= 1 of the bound extremes
        p1 = np.array([-100,100,-99,80,95.0,-100])
        for i in range(100):
            p2 = numerical_generator.random_npoint_from_point(p1,r)
            assert type(p2) != type(None)
        return

    def test__CycleMap__next(self): 
        cm = numerical_generator.CycleMap(5)
        d = OrderedDict({10:13,13:74,74:21,21:16,16:10})
        cm.set_map(d)
        q = set() 

        for i in range(10): 
            q_ = next(cm) 
            q |= {q_}
        assert len(q) == 5 

        cm = numerical_generator.CycleMap(7)
        f = numerical_generator.CycleMap.random_cycle_map(7)
        cm.set_map(f) 
        q = set()
        for i in range(10): 
            q_ = next(cm) 
            q |= {q_}
        assert len(q) == 7

    def test__prg_seqsort_ties(self): 
        prg0 = numerical_generator.prg__constant(5)
        prg4 = numerical_generator.prg__n_ary_alternator(s0=3,s1=25,start=1) 
        R2 = ["3","52432","cat","DGO","bingo","ROGUE", "rauchu", "NOBODIES","12,55","32","43,10,13","12"] 
        Q = [(40,r) for r in R2] 
        vf = lambda x: x[1] 
        rq = numerical_generator.prg_seqsort_ties(Q,prg4,vf) 

        assert rq == [(40, '12'), (40, '12,55'), (40, '3'), \
            (40, '32'), (40, '43,10,13'), (40, '52432'), \
            (40, 'DGO'), (40, 'NOBODIES'), (40, 'ROGUE'), \
            (40, 'bingo'), (40, 'cat'), (40, 'rauchu')]

        Q2 = [(prg4(),r) for r in R2] 
        Q2.extend([(prg4(),r) for r in R2]) 
        Q2.extend([(prg4(),r) for r in R2]) 

        vf2 = lambda x: x[0] 
        rq2 = numerical_generator.prg_seqsort_ties(Q2,prg4,vf2) 
        rq3 = numerical_generator.prg_seqsort_ties(Q2,prg0,vf2) 

        assert rq2 == [(3, '32'), (3, 'NOBODIES'), \
            (4, '43,10,13'), (4, '12,55'), (5, '12'), \
            (5, '32'), (6, '3'), (6, '43,10,13'), \
            (7, '52432'), (7, '12'), (8, 'cat'), \
            (9, 'DGO'), (10, 'bingo'), (11, 'ROGUE'), \
            (12, 'rauchu'), (13, 'NOBODIES'), (14, '12,55'), \
            (15, '32'), (16, '3'), (16, '43,10,13'), \
            (17, '52432'), (17, '12'), (18, 'cat'), (18, '3'), \
            (19, 'DGO'), (19, '52432'), (20, 'bingo'), (20, 'cat'), \
            (21, 'ROGUE'), (21, 'DGO'), (22, 'rauchu'), \
            (22, 'bingo'), (23, 'NOBODIES'), (23, 'ROGUE'), \
            (24, '12,55'), (24, 'rauchu')]

        assert rq3 == [(3, 'NOBODIES'), (3, '32'), (4, '12,55'), \
            (4, '43,10,13'), (5, '32'), (5, '12'), (6, '43,10,13'), \
            (6, '3'), (7, '12'), (7, '52432'), (8, 'cat'), (9, 'DGO'), \
            (10, 'bingo'), (11, 'ROGUE'), (12, 'rauchu'), \
            (13, 'NOBODIES'), (14, '12,55'), (15, '32'), \
            (16, '43,10,13'), (16, '3'), (17, '12'), (17, '52432'), \
            (18, '3'), (18, 'cat'), (19, '52432'), (19, 'DGO'), \
            (20, 'cat'), (20, 'bingo'), (21, 'DGO'), (21, 'ROGUE'), \
            (22, 'bingo'), (22, 'rauchu'), (23, 'ROGUE'), (23, 'NOBODIES'), \
            (24, 'rauchu'), (24, '12,55')]

    def test__prg_partition_for_sz(self): 
        import random 
        random.seed(100)

        lc = numerical_generator.LCG(32,3901)
        def prg1_(): 
            def f(): 
                return next(lc) 
            return f 

        prg1 = prg1_() 
        S = 100
        num_sets = 100 
        p1 = numerical_generator.prg_partition_for_sz(S,num_sets,prg1,0.0)
        assert p1 == [1] * 100 

        num_sets = 50 
        p2 = numerical_generator.prg_partition_for_sz(S,num_sets,prg1,0.0)
        assert p2 == [2] * 50

        prg1 = prg1_() 
        num_sets = 50 
        p3 = numerical_generator.prg_partition_for_sz(S,num_sets,prg1,0.5)
        assert p3 == [30, 18, 4, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
            1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
            1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
            1, 1, 1],"got {}".format(p3)
        assert sum(p3) == 100 


        num_sets = 50 
        p4 = numerical_generator.prg_partition_for_sz(S,num_sets,prg1,0.1)
        assert p4 == [31, 3, 8, 12, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
            1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
            1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, \
            1, 1, 1],"got {}".format(p4) 
        assert sum(p4) == 100 

        num_sets = 15 
        p5 = numerical_generator.prg_partition_for_sz(S,num_sets,prg1,0.1)
        assert p5 == [56, 27, 5, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], \
            "got {}".format(p5) 
        assert sum(p5) == 100 

        num_sets = 4
        p6 = numerical_generator.prg_partition_for_sz(S,num_sets,prg1,0.1)
        assert p6 == [86, 2, 9, 3], "got {}".format(p6) 
        assert sum(p6) == 100 

        prng = numerical_generator.prg__LCG(-4,5,2,29)
        S = 100
        num_sets = 20
        prt = numerical_generator.prg_partition_for_sz(S,num_sets,prng,0.0)
        assert prt == [5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 26]
        assert sum(prt) == 100 



if __name__ == '__main__':
    unittest.main()
